module CommonSyntax where

import List(intersperse)
import Pretty

-- A class for pretty-printing

class Pretty p where pretty :: a -> p -> Doc

---------- pretty printing utilities -------------------------------------------------------------

vcatP :: Pretty b => a -> [b] -> Doc
vcatP s = vcat . map (pretty s)

sepP :: Pretty b => a -> [b] -> Doc
sepP s = sep . map (pretty s)

sepByCommas :: Pretty b => a -> [b] -> Doc
sepByCommas s = hcat . intersperse (text ", ") . map (pretty s)

parenCommas :: Pretty b => a -> [b] -> Doc
parenCommas s = parens . sepByCommas s


data SrcLoc = SrcLoc Int Int   -- (Line, Indentation)


data MatchStrategy = FunctionalMatch | OptMatch
     deriving (Eq)

instance Pretty MatchStrategy where
   pretty _ FunctionalMatch = text "funcmatch"   -- SUP: Never used in the end
   pretty _ OptMatch        = text "optmatch"


data Fixity = Fixity Precedence Associativity

instance Pretty Fixity where
   pretty s (Fixity p a) = pretty s p <+> pretty s a


newtype Precedence = Precedence Int   -- between 0 and 9

instance Pretty Precedence where
   pretty _ (Precedence p) = int p


data Associativity = LeftAssoc | RightAssoc | NonAssoc

instance Pretty Associativity where
   pretty _ LeftAssoc  = text "infixl"
   pretty _ RightAssoc = text "infixr"
   pretty _ NonAssoc   = text "infix"


data Annotation = Flex | Rigid | Choice | DefaultAnnotation   -- SUP: internal use only
       deriving (Show,Ord,Eq) -- for debugging purposes only


instance Pretty Annotation where
   pretty _ Flex              = text "flex"
   pretty _ Rigid             = text "rigid"
   pretty _ Choice            = text "choice"
   pretty _ DefaultAnnotation = text "defaultAnnotation"


data Literal = LiteralInt Int
     deriving (Show, Eq, Ord)

instance Pretty Literal where
   pretty _ (LiteralInt i) = int i
