module Compiler(compile)
where

import CurryCoreSyntax
import InstructionSet
import CommonSyntax

-- TODO suchvariablen

compile :: CCProgram -> Instructions
------------------------------------
compile = compileExpr False 0




compileExpr b n (CCExprLet fundefs expr) =
   (LET (length fundefs) (map compileFunDef fundefs)) : (compileExpr b n expr)
compileExpr b n (CCExprFreeDecl _ expr) =
   LETFREE : (compileExpr b n expr)
compileExpr b n (CCExprLambda (CCVariable id) expr) =
   if (b || n > 0) then
      (TAKE id) : (compileExpr b (n-1) expr)
   else
      (VALUE) : (TAKE id) : (compileExpr True n expr)
compileExpr b n (CCExprCase Rigid var alts) =
   (EVAL $ compileVar var):WAIT:(compileSwitch b n alts)
compileExpr b n (CCExprCase Flex var alts) =
   (EVAL $ compileVar var):(BOUND $ compileSwitch b n alts):(compileIndet b n alts)
compileExpr b n (CCExprAppl expr (CCVariable id)) =
   (ARG id):(compileExpr b (n+1) expr)
compileExpr b n (CCExprVar var) =
   compileVar var
compileExpr b n (CCExprConstr c vars) =
   compileConstr c vars
--   (VALUE $ compileConstr c vars) : [RETURN]
compileExpr b n (CCExprStrict expr) =
   (EVAL $ compileExpr b n expr) : [STRICT, RETURN] -- HALT statt RETURN?
compileExpr b n (CCExprAnd expr1 expr2) =
   (A_THREADS $ compileExpr b n expr1) : (compileExpr b n expr2)
compileExpr b n (CCExprOr expr1 expr2) =
   (TRY $ compileExpr b n expr1) : (TRY $ compileExpr b n expr2) : [BACKTRACK]
compileExpr b n (CCExprUnify expr1 expr2) =
   (U_THREADS $ compileExpr b n expr1) : (compileExpr b n expr2)
compileExpr b n (CCExprGuard expr1 expr2) =
   (EVAL $ compileExpr b n expr1) : compileExpr b n expr2
compileExpr b n CCExprFail =
   [BACKTRACK]
compileExpr b n (CCExprError error) =
   compileError error


compileFunDef (CCFunDef (CCVariable id) expr) =
   (EVAL $ compileExpr False 0 expr) : [UPDATE id, DEREF id]
   

compileSwitch b n alts =
   let compileSwitch' b n (CCAlternative (CCDataConstr _ id _) vars expr) =
          (map (\ (CCVariable id) -> (BIND id)) vars) ++ compileExpr b n expr
   in  [SWITCH $ map (compileSwitch' b n) alts]

compileIndet b n alts =
   WAKE_UP_THREADS : ((map (compileTry b n) alts) ++ [BACKTRACK])


compileTry b n (CCAlternative c vars expr) =
   TRY $ (map (\ _ -> LETFREE) vars) ++ [BIND_TO $ compileConstr c vars] ++ compileExpr b n expr


compileVar (CCVariable id) =
   [DEREF id, UVAR id] -- : [UPDATE id, RETURN]
--   (DEREF id [VALUE [UVAR id], RETURN]) : [UPDATE id, RETURN]


compileConstr (CCDataConstr (CCName n) id _) vars =
   ((CONSTR id (compilePattern vars) n):compileStrict vars)

compilePattern vars =
   concatMap compileVar vars

compileStrict []   = [RETURN]
compileStrict vars = 
   let compileStrictV (CCVariable id) =
            (EVAL [DEREF id, UVAR id]) : [STRICT, RETURN]
--             (DEREF id [UVAR id]) : [UPDATE id, STRICT, RETURN]
--             (DEREF id [VALUE [UVAR id], RETURN]) : [UPDATE id, STRICT, RETURN]
   in  (map (\ v -> A_THREADS $ compileStrictV v) (tail vars)) ++ compileStrictV (head vars)


compileError CCErrorPatternDecl = [error "pattern declararation not fitted!\n"] -- ??