module DispatcherPro
where

import Queue
import Stack(emptyStack)
import InstructionSet
import StateComponents
import FiniteMap
import List



rNext r = dequeueWithDefault r 
			     (error "No more threads: deadlock or blackhole occured!\n")

sLookup s p = lookupWithDefaultFM s [] p

tLookup t d = lookupWithDefaultFM t 
				  (error "Dispatcher.tLookup: Current thread not initialized in thread tree!\n")
				  d

cLookup c d = lookupWithDefaultFM c 
				  (error "Dispatcher.cLookup: Current thread not initialized in content!\n")
				  d





newThreads :: Dispatcher -> Closure -> Closure -> Dispatcher
------------------------------------------------------------
newThreads (Dispatcher f r s t c st d) c1 c2 =
   let (d1:d2:f')   = f
       r'           = enqueue (enqueue r d1) d2
       (_,_,df)	    = tLookup t d
       t'	    = addListToFM t [(d, (d1,d2,df)), 
				     (d1,(0, 0, d )), 
				     (d2,(0, 0, d ))]
       c'	    = addListToFM c [(d1, Thread c1 emptyClosure emptyStack emptyStack), 
				     (d2, Thread c2 emptyClosure emptyStack emptyStack)]
   in  Dispatcher f' r' s t' c' st d


newNotRunnableThreads :: Dispatcher -> Dispatcher
-------------------------------------------------
newNotRunnableThreads (Dispatcher f r s t c st d) =
   let (d1:d2:f')   = f
       (_,_,df)	    = tLookup t d
       t'	    = addListToFM t [(d, (d1,d2,df)), 
				     (d1,(0, 0, d )), 
				     (d2,(0, 0, d ))]
       c'	    = addListToFM c [(d1, Thread emptyClosure emptyClosure emptyStack emptyStack), 
				     (d2, Thread emptyClosure emptyClosure emptyStack emptyStack)]
   in  Dispatcher f' r s t' c' st d


actualizeThread :: Dispatcher -> Thread -> Dispatcher
-----------------------------------------------------
actualizeThread (Dispatcher f r s t c st d) thread =
   let c'      = addToFM c d thread
   in  (Dispatcher f r s t c' st d)


nextThread :: Dispatcher -> (Dispatcher, Thread)
------------------------------------------------
nextThread (Dispatcher f r s t c st _) =
   let (r',d') = rNext r
   in  (Dispatcher f r' s t c st d', cLookup c d')


childsTerminated :: Dispatcher -> Bool
--------------------------------------
childsTerminated (Dispatcher _ _ _ _ _ st d) = 
   d `elem` st


removeChilds :: Dispatcher -> Dispatcher
----------------------------------------
removeChilds (Dispatcher f r s t c st d) =
   let (ds1, ds2, _)  = tLookup t d
       t'	      = delListFromFM t [ds1, ds2]
       c'	      = delListFromFM c [ds1, ds2]
       st'	      = st \\ [d]
       f'	      = ds1:ds2:f
   in  Dispatcher f' r s t' c' st' d


actualizeThreadsValue :: Dispatcher -> Closure -> Dispatcher
------------------------------------------------------------
actualizeThreadsValue (Dispatcher f r s t c st d) v =
   let Thread cl _ a' r' = cLookup c d
       c'	         = addToFM c d (Thread cl v a' r')
   in  (Dispatcher f r s t c' st d)



lSonThread :: Dispatcher -> (Dispatcher, Thread)
------------------------------------------------
lSonThread (Dispatcher f r s t c st d) =
   let (dl,_,_)         = tLookup t d
   in  (Dispatcher f r s t c st dl, cLookup c dl)


rSonThread :: Dispatcher -> (Dispatcher, Thread)
------------------------------------------------
rSonThread (Dispatcher f r s t c st d) =
   let (_,dr,_)         = tLookup t d
   in  (Dispatcher f r s t c st dr, cLookup c dr)


parentThread :: Dispatcher -> (Dispatcher, Thread)
--------------------------------------------------
parentThread (Dispatcher f r s t c st d) =
   let (_,_, df)         = tLookup t d
   in  (Dispatcher f r s t c st df, cLookup c df)


firstChildTerminated :: Dispatcher -> Dispatcher
------------------------------------------------
firstChildTerminated (Dispatcher f r s t c st d) =
   Dispatcher f r s t c (d:st) d



childValues :: Dispatcher -> (Closure, Closure)
-----------------------------------------------
childValues (Dispatcher _ _ _ t c _ d) =
   let (ds1, ds2, _)    = tLookup t d
       Thread _ v1 _ _  = cLookup c ds1
       Thread _ v2 _ _  = cLookup c ds2
   in  (v1, v2)



suspendThread :: Dispatcher -> HeapAddr -> Dispatcher
-----------------------------------------------------
suspendThread (Dispatcher f r s t c st d) p =
   let s'      = addToFM s p (d:(sLookup s p))
   in  Dispatcher f r s' t c st d



wakeUpThreads :: Dispatcher -> HeapAddr -> Dispatcher
-----------------------------------------------------
wakeUpThreads (Dispatcher f r s t c st d) p =
   let threadIds = sLookup s p
       s'	 = delFromFM s p
       r'	 = foldr (\ d r -> enqueue r d) r threadIds
   in  Dispatcher f r' s' t c st d


runnableThreadsLeft :: Dispatcher -> Bool
-----------------------------------------
runnableThreadsLeft (Dispatcher _ r _ _ _ _ _) =
   not $ isQueueEmpty r


parentThreadExists :: Dispatcher -> Bool
----------------------------------------
parentThreadExists (Dispatcher _ _ _ t _ _ d) =
   let (_,_,df) = tLookup t d 
   in  df /= -1