module Flatten(  c_and_addr
	       , c_unify_addr
	       , c_start_addr
	       , cIdOffset
	       , cArityOffset
	       , cStrictAddrOffset
	       , cPatternOffset
	       , nextVarOffset
	       , flatten )
where


import MyPrelude
import InstructionSet
import TraceClass
import Array




-- basic code sequences

c_and   = [NF_AWAIT_SONS, NF_TERM_THREADS, NF_CONSTANT "Success" (-1), NF_RETURN]
c_unify = [NF_AWAIT_SONS, NF_UNIFY,	   NF_CONSTANT "Success" (-1), NF_RETURN]

-- basic addresses

c_and_addr    :: CAddr
c_and_addr    = 0
c_unify_addr  = flatInstrsLen c_and
c_start_addr  = c_unify_addr + flatInstrsLen c_unify 


-- some offsets

cIdOffset         :: Int	
cIdOffset         =  1
cArityOffset      :: Int
cArityOffset      =  2
cStrictAddrOffset :: Int
cStrictAddrOffset =  3
cPatternOffset	  :: Int
cPatternOffset	  =  4

nextVarOffset	  :: Int
nextVarOffset	  =  4



flatten :: NF_Instructions -> Code
----------------------------------
flatten code =
   let is = (fst $ (unAddrM (flatInstructions (c_and ++ c_unify ++ code))) 0)
   in  mkCode is
	     




flatInstructions :: NF_Instructions -> AddrM [Instruction]
----------------------------------------------------------


flatInstructions [] = return []

-- first the flat instructions
flatInstructions (NF_LETFREE:r)               = do incAddr
						   rest <- flatInstructions r
						   return $ [LETFREE] ++ rest
flatInstructions ((NF_ARG a):r)		      = do incAddrBy 2
						   rest <- flatInstructions r
						   return $ [ARG, INT a] ++ rest
flatInstructions ((NF_TAKE a):r)	      = do incAddrBy 2
						   rest <- flatInstructions r
						   return $ [TAKE, INT a] ++ rest
flatInstructions ((NF_DEREF a):r)	      = do incAddrBy 2
						   rest <- flatInstructions r
						   return $ [DEREF, INT a] ++ rest
flatInstructions ((NF_RETURN):r)	      = do incAddr
						   rest <- flatInstructions r
						   return $ [RETURN] ++ rest
flatInstructions (NF_VALUE:r)		      = do incAddr
						   rest <- flatInstructions r
						   return $ [VALUE] ++ rest
flatInstructions ((NF_UPDATE a):r)	      = do incAddrBy 2
						   rest <- flatInstructions r
						   return $ [UPDATE, INT a] ++ rest
flatInstructions (NF_STRICT:r)		      = do incAddr
						   rest <- flatInstructions r
						   return $ [STRICT] ++ rest
flatInstructions ((NF_UVAR a):r)	      = do incAddrBy 2
						   rest <- flatInstructions r
						   return $ [UVAR, INT a] ++ rest
flatInstructions ((NF_CONSTANT n a):r)	      = do incAddrBy 2
						   rest <- flatInstructions r
						   return $ [CONSTANT n, INT a] ++ rest
flatInstructions (NF_WAIT:r)		      = do incAddr
						   rest <- flatInstructions r
						   return $ [WAIT] ++ rest
flatInstructions (NF_BACKTRACK:r)	      = do incAddr
						   rest <- flatInstructions r
						   return $ [BACKTRACK] ++ rest
flatInstructions (NF_BIND:r)		      = do incAddr
						   rest <- flatInstructions r
						   return $ [BIND] ++ rest
flatInstructions (NF_UNIFY:r)		      = do incAddr
						   rest <- flatInstructions r
						   return $ [UNIFY] ++ rest
flatInstructions (NF_WAKE_UP_THREADS:r)	      = do incAddr
						   rest <- flatInstructions r
						   return $ [WAKE_UP_THREADS] ++ rest
flatInstructions (NF_TERM_THREADS:r)	      = do incAddr
						   rest <- flatInstructions r
						   return $ [TERM_THREADS] ++ rest
flatInstructions (NF_AWAIT_SONS:r)	      = do incAddr
						   rest <- flatInstructions r
						   return $ [AWAIT_SONS] ++ rest
flatInstructions ((NF_BLACKHOLE_MARKER a):r)  = do incAddrBy 2
						   rest <- flatInstructions r
						   return $ [BLACKHOLE_MARKER, INT a] ++ rest
flatInstructions ((NF_ERROR e):r)	      = do incAddr
						   rest <- flatInstructions r
						   return $ [ERROR e] ++ rest



-- no the ugly stuff

-- NF_LET n [is1...isn] --> LET n addr1...addrn ...
flatInstructions ((NF_LET n isList):r) =
   do incAddrBy (n+2)
      rest <- flatInstructions r
      addr <- getAddr
      args <- concatMapM flatInstructions isList
      let argLength = [flatInstrsLen is | is <- isList]
	  offsets   = accumulator 0 argLength
	  addrs = [INT (addr + offset) | offset <- offsets]
      return $ [LET, INT n] ++ addrs ++ rest ++ args



flatInstructions ((NF_EVAL is):r) =
   do incAddrBy 2
      arg  <- flatInstructions is
      addr <- getAddr
      rest <- flatInstructions r
      return $ [EVAL, INT addr] ++ arg ++ rest




flatInstructions ((NF_CONSTR name id arity is):r) =
   do incAddrBy 4
      pattern_code <- flatInstructions is
      addr	   <- getAddr
      strict_code  <- flatInstructions r
      return $ [CONSTR name, INT id, INT arity, INT addr] ++ pattern_code ++ strict_code



-- switch [alt1 ... altn] -> switch addr1...addrn alt1...altn  ...
flatInstructions ((NF_SWITCH alts):r) =
   let altLength = [flatInstrsLen is | is <- alts]
       n	 = length alts
       offsets	 = accumulator 0 altLength
   in  do incAddrBy (n + 1)
	  addr <- getAddr
 	  let addrs = [INT (addr + offset) | offset <- offsets]
 	  flatAlts <- concatMapM flatInstructions alts
 	  rest	   <- flatInstructions r -- should be empty
 	  return $ [SWITCH] ++ addrs ++ flatAlts ++ rest


flatInstructions ((NF_TRY is):r) =
   do incAddrBy 2
      try  <- flatInstructions is
      addr <- getAddr
      alt  <- flatInstructions r
      return $ [TRY, INT addr] ++ try ++ alt





flatInstructions ((NF_BOUND is):r) =
   do incAddrBy 2
      bound_code  <- flatInstructions is
      addr	  <- getAddr
      ubound_code <- flatInstructions r
      return $ [BOUND, INT addr] ++ bound_code ++ ubound_code






flatInstructions ((NF_BIND_TO is):r) =
   do incAddrBy 2
      rest    <- flatInstructions r
      addr    <- getAddr
      binding <- flatInstructions is
      return $ [BIND_TO, INT addr] ++ rest ++ binding





flatInstructions ((NF_A_THREADS is):r) =
   do incAddrBy 2
      thread1 <- flatInstructions is
      addr    <- getAddr
      thread2 <- flatInstructions r
      return $ [A_THREADS, INT addr] ++ thread1 ++ thread2





flatInstructions ((NF_U_THREADS is):r) =
   do incAddrBy 2
      thread1 <- flatInstructions is
      addr    <- getAddr
      thread2 <- flatInstructions r
      return $ [U_THREADS, INT addr] ++ thread1 ++ thread2




-- AddrM --
-----------

newtype AddrM a = AddrM (CAddr -> (a, CAddr))


instance Monad AddrM where
   m >>= k    = AddrM (\ addr -> 
			   let (val, addr') = (unAddrM m) addr
			   in  unAddrM (k val) addr')
   return val = AddrM (\ addr -> (val, addr))


unAddrM (AddrM m) = m

incAddr =
   AddrM (\ a -> ((), a+1))

incAddrBy n =
   AddrM (\ a -> ((), a+n))

getAddr =
   AddrM (\ a -> (a, a))




-- flatInstrsLen --
-------------------


flatInstrsLen  = sum . map flatInstrLen

flatInstrLen (NF_LETFREE)            = 1
flatInstrLen (NF_ARG _)		     = 2
flatInstrLen (NF_TAKE _)	     = 2
flatInstrLen (NF_DEREF _)	     = 2
flatInstrLen (NF_RETURN)	     = 1
flatInstrLen (NF_VALUE)		     = 1
flatInstrLen (NF_UPDATE _)	     = 2
flatInstrLen (NF_STRICT)	     = 1
flatInstrLen (NF_UVAR _)	     = 2
flatInstrLen (NF_CONSTANT _ _)	     = 2
flatInstrLen (NF_WAIT)		     = 1
flatInstrLen (NF_BACKTRACK)	     = 1
flatInstrLen (NF_BIND)		     = 1
flatInstrLen (NF_UNIFY)		     = 1
flatInstrLen (NF_WAKE_UP_THREADS)    = 1
flatInstrLen (NF_TERM_THREADS)	     = 1
flatInstrLen (NF_AWAIT_SONS)	     = 1
flatInstrLen (NF_BLACKHOLE_MARKER _) = 2
flatInstrLen (NF_ERROR _)	     = 1

flatInstrLen (NF_LET n isList)	     = 2 + n + (sum $ map flatInstrsLen isList)
flatInstrLen (NF_EVAL is)	     = 2 + flatInstrsLen is
flatInstrLen (NF_CONSTR _ _ _ is)    = 4 + flatInstrsLen is
flatInstrLen (NF_SWITCH alts)	     = 1 + (length alts) + (sum $ map flatInstrsLen alts)
flatInstrLen (NF_TRY is)	     = 2 + flatInstrsLen is
flatInstrLen (NF_BOUND is)	     = 2 + flatInstrsLen is
flatInstrLen (NF_BIND_TO is)	     = 2 + flatInstrsLen is
flatInstrLen (NF_A_THREADS is)	     = 2 + flatInstrsLen is
flatInstrLen (NF_U_THREADS is)	     = 2 + flatInstrsLen is







