module InstructionSet
where

import MyPrelude
import TraceClass
import Array

type NF_Instructions = [NF_Instruction]
type Instructions    = [Instruction]


type CAddr = Int
data Code  = Code (Array CAddr Instruction)


instance Show Code where
   showsPrec _ (Code code) = 
      let pairs = assocs code
      in  showChar '\t' . showsSepBy "\n\t" pairs


instance Trace Code where
   trace c@(Code code) =
      do putStrLn $ "Code address range: " ++ show (bounds code) 
	 putStrLn "Which code address do you want to inspect?"
	 putStr "-> "
	 addr <- getInt
	 traceCode c addr

traceCode code addr =
   let res = lookupCodeMB code addr
   in  case res of
	Nothing -> putStr "Inaccessible address."
	Just i  -> do putStrLn $ "Content of address " ++ show addr ++ ":"
		      putStrLn $ show i
		      case i of
			LET	      -> do let i'@(INT n) = lookupCode code (addr+1)
					    putStrLn $ show i'
					    mapM_ (putStrLn . show) [lookupCode code (addr + offset) |  offset <- [2..n+1]]
			ARG	      ->  putStrLn $ show (lookupCode code (addr+1))
			TAKE	      ->  putStrLn $ show (lookupCode code (addr+1))
			EVAL	      ->  putStrLn $ show (lookupCode code (addr+1))
			DEREF	      ->  putStrLn $ show (lookupCode code (addr+1))
			UPDATE	      ->  putStrLn $ show (lookupCode code (addr+1))
			CONSTR _      ->  do putStrLn $ show (lookupCode code (addr+1))
					     putStrLn $ show (lookupCode code (addr+2))
					     putStrLn $ show (lookupCode code (addr+3))
			UVAR	      ->  putStrLn $ show (lookupCode code (addr+1))
			CONSTANT _    ->  putStrLn $ show (lookupCode code (addr+1))
			TRY	      ->  putStrLn $ show (lookupCode code (addr+1))
			BOUND	      ->  putStrLn $ show (lookupCode code (addr+1))
			BIND_TO	      ->  putStrLn $ show (lookupCode code (addr+1))
			A_THREADS     ->  putStrLn $ show (lookupCode code (addr+1))
			U_THREADS     ->  putStrLn $ show (lookupCode code (addr+1))
			BLACKHOLE_MARKER     -> putStrLn $ show (lookupCode code (addr+1))
			_		     -> return () 


mkCode :: Instructions -> Code
------------------------------
mkCode is = Code $ listArray (0,((length is) - 1)) is 


lookupCode (Code code) addr = code!addr
		   

lookupCodeMB (Code code) addr =
   let (lower, upper) = bounds code
   in  if (addr < lower || addr > upper) then
	  Nothing
        else Just (code!addr)



data Instruction =
       INT		Int
     | LET
     | LETFREE
     | ARG
     | TAKE
     | EVAL
     | DEREF
     | RETURN
     | VALUE
     | UPDATE
     | STRICT
     -- terms
     | CONSTR	        String -- name necessary, as we have no type informations :-(
     | UVAR
     | CONSTANT         String -- dito
     -- pattern matching
     | WAIT
     | SWITCH
     | TRY	
     | BACKTRACK
     | BOUND	
     | BIND	
     | BIND_TO	
     -- unification
     | UNIFY
     -- threads
     | A_THREADS
     | U_THREADS
     | WAKE_UP_THREADS
     | TERM_THREADS
     | AWAIT_SONS
     ---
     | BLACKHOLE_MARKER
     -- error
     | ERROR		String -- would be implemented by a table
     deriving (Eq,Show)

data NF_Instruction =
       NF_LETFREE
     | NF_ARG		Int
     | NF_TAKE		Int
     | NF_DEREF	        Int
     | NF_RETURN
     | NF_VALUE
     | NF_UPDATE	Int
     | NF_STRICT
     -- terms
     | NF_UVAR	        Int
     | NF_CONSTANT      String Int
     -- pattern matching
     | NF_WAIT
     | NF_BACKTRACK
     | NF_BIND
     -- unification
     | NF_UNIFY
     -- threads
     | NF_WAKE_UP_THREADS
     | NF_TERM_THREADS
     | NF_AWAIT_SONS
     ---
     | NF_BLACKHOLE_MARKER Int
     --- really not flat instructions
     | NF_LET           Int            [NF_Instructions]
     | NF_EVAL		NF_Instructions
     | NF_CONSTR	String          Int     Int       NF_Instructions   -- name necessary as we have no type informations :-(
     | NF_SWITCH       [NF_Instructions]
     | NF_TRY	        NF_Instructions
     | NF_BOUND	        NF_Instructions
     | NF_BIND_TO	NF_Instructions
     | NF_A_THREADS     NF_Instructions
     | NF_U_THREADS	NF_Instructions
     -- error
     | NF_ERROR		String
     deriving (Show)

instance Trace NF_Instruction where
   trace i = putStr $ show i -- TODO

instance Trace Instruction where
   trace i = putStr $ show i -- TODO