module Main 
where


import MyPrelude
import ErrorMonad
import System(getArgs)
import GetOpt

import Pretty(render, empty, (<+>), (<>), int, text, parens)

import CommonSyntax
import Parser(parseModule)
import MangleRdrToTc
import MangleTcToCC
import Compiler
import Flatten
import Machine


----------------------------------------------------------------------


data Option  = ShowReader 
	     | ShowTc
	     | ShowCC
	     | ShowSearchTab
	     | ShowCode
	     | ShowFlat
	     | Quiet
	     | TraceInterpretation
	     ----------------------
	     | Help
	     deriving (Eq)

options :: [OptDescr Option]
options =
   [Option ['r'] ["show-reader"]         (NoArg ShowReader)           "show reader result",
    Option ['t'] ["show-type-checker"]	 (NoArg ShowTc)		      "show type checker input",
    Option ['c'] ["show-curry-core"]     (NoArg ShowCC)		      "show curry core program",
    Option ['v'] ["show-search-var"]	 (NoArg ShowSearchTab)        "show search variables",    
    Option ['i'] ["show-instructions"]   (NoArg ShowCode)             "show machine code",    
    Option ['f'] ["show-flat-code"]	 (NoArg ShowFlat)             "show flattened machine code",    
    Option ['q'] ["quiet"]		 (NoArg Quiet)		      "do not interprete",
    Option ['s'] ["step"]		 (NoArg TraceInterpretation)  "interprete stepwise",
    Option ['h'] ["help"]                (NoArg Help)                 "describe options"]


usage :: IO ()
usage = putStr (usageInfo ("\nThis is the MAchine for Curry (MAC)" ++
			   "\nUsage: mac [OPTION...] filename\nwhere OPTION is one of:") options)

main :: IO ()
main = do args <- getArgs
          mainHugs args


testT = mainHugs ["examples/myTest.curry", "-s", "-c", "-v"]
test  = mainHugs ["examples/myTest.curry"]


mainHugs :: [String] -> IO ()
mainHugs args = handleCmdline (getOpt Permute options args)
   where handleCmdline (opts,fileNames,[]  ) | Help `elem` opts = usage
                                             | otherwise        = go opts fileNames
         handleCmdline (_   ,_        ,errs)                    = do putStr (concat errs)
                                                                     usage


go options fileNames =  do let isOpt x =  x `elem` options
			       out msg what = do putStrLn ("---------- " ++ msg ++" ---------")
						 putStrLn . render . pretty undefined $ what
			   input <- case fileNames of
					 []  -> getContents
					 [f] -> readFile f
					 fs  -> ioError (userError ("expecting max. 1 file name: " ++ show fs))
			   -- parser
			   pm	 <- maybeErrorToIO (parseModule input)
			   ioIf (isOpt ShowReader) (out "Reader" pm)
			   -- mangle to type checker syntax
			   tm <- maybeErrorToIO (rdrModuleToTcModule FunctionalMatch Rigid pm)
			   ioIf (isOpt ShowTc) (out "type checker " tm)
			   -- transform curry_s to curry_core
			   (cc,searchtab) <- maybeErrorToIO (mangleTcToCC tm)
			   ioIf (isOpt ShowCC) (out  "curry core" cc)
			   ioIf (isOpt ShowSearchTab) 
				(putStr $ "main:\t\t\t" ++ show (fst searchtab) ++ "\nSearch variables:\t" ++ show (snd searchtab) ++ "\n")
			   -- translate curry_core to machine code
			   let code = compile cc
			   ioIf (isOpt ShowCode) (putStr $ "\n" ++ show code ++ "\n")
			   -- flatten the machine code
			   let fcode = flatten code
			   ioIf (isOpt ShowFlat) (putStr $ "\n" ++ show fcode ++ "\n")
			   -- interprete the generated code
			   ioIf (not (isOpt Quiet)) (interprete searchtab fcode (isOpt TraceInterpretation))





