module MangleTcToCC(mangleTcToCC, SearchVarTab)
where

--import Trace(trace)

import MyPrelude

import CommonSyntax
import TypeCheckerSyntax
import PrettyTypeCheckerSyntax
import CurryCoreSyntax
import PrettyCurryCoreSyntax
import Pretty(render, empty, (<+>), (<>), int, text, parens)
import ErrorMonad

import List
import Maybe
import Monad
import FiniteMap


-- TODO: Die Behandlung von Literalen ist offensichtlich
--       unzureichend

-- we assume (additionaly to what is assured already):
--
-- - the program to be well typed
-- - saturation of constructor applications in patterns
--
-- tasks:
--
-- - every variable defined ?						     DONE
-- - every constructor defined ?					     DONE
-- - assure that the evaluation strategie choice                             IMPOSSIBLE
--   does not occur							     ----
-- - assure no anonymous variables are used                                  DONE
-- - transform literals to constructors					     DONE
-- - handle anonymous variables						     DONE
-- - transform let / letrec                                                  DONE
-- - assure no qualified names are used					     DONE
-- - remove the distinction between function names and variable names	     DONE
-- - make builtin operations explicit					     DONE
-- and, finally, provide the transformations described in the thesis, i.e.   ----
-- - Transform program to expression					     DONE
-- - assure existence of main						     DONE
-- - transform guards							     DONE
-- - letfree								     DONE
-- - not saturated constructor applications (except constants)               DONE
-- - arguments must be variables only					     DONE
-- - tDTree								     DONE
-- - tLambda								     DONE
-- - tPDecl								     DONE
-- here, the following points are already cleared bei SUP:-----------------------
-- - whereDefs are already eliminated
-- - lists are already transformed to constructor applications
-- (- if then else statements are already transformed)
--
-- differences to thesis:
-- - we have no type information :-( 
--   therefore all guards are assumed to be constraints
-- - ifthenelse / Bool are assumed to be defined in the sourcecode (Prelude)



defaultAnnot = Rigid
--             HACK!



builtins = ["=:=", "_guard", "&"]

freeLDefs ldefs			                 = (nub $ concatMap freeLDef ldefs) \\ definedByLDefs ldefs
freeLDef (TcLocalFuncDecl _ _ (TcName v) equs)   = freeEqus equs \\ [v] 
freeLDef (TcLocalPatternDecl pattern expr)       = freeExpr expr \\ freePattern pattern
freeEqus equs					 = nub $ concatMap freeEqu equs
freeEqu (TcEquat patterns expr)			 = freePatterns patterns `union` freeExpr expr
freePatterns patterns				 = nub $ concatMap freePattern patterns
freePattern (TcPatternVar (TcName v))		 = [v]
freePattern (TcPatternAppl _ patterns)		 = freePatterns patterns
freePattern _					 = []
freeExpr (TcExprLambda patterns expr)		 = freeExpr expr  \\ freePatterns patterns
freeExpr (TcExprFreeDecl vars expr)		 = freeExpr expr  \\ [v | (TcName v) <- vars]
freeExpr (TcExprLet ldef expr)			 = (freeExpr expr \\ definedByLDef  ldef)  `union` freeLDef  ldef
freeExpr (TcExprLetrec ldefs expr)		 = (freeExpr expr \\ definedByLDefs ldefs) `union` freeLDefs ldefs
freeExpr (TcExprAppl expr1 expr2)		 = freeExpr expr1 `union` freeExpr expr2
freeExpr (TcExprVar (TcUnqualName v))		 = [v] \\ builtins
freeExpr _					 = []
definedByLDef (TcLocalFuncDecl _ _ (TcName v) _) = [v]
definedByLDef (TcLocalPatternDecl pattern _)	 = freePattern pattern
definedByLDefs ldefs				 = concatMap definedByLDef ldefs



-- This is the hack for a "%="!*#+ curry feature
superfeature localdefs =
   let arity (TcLocalFuncDecl _ _ _ ((TcEquat ps _):_)) = length ps
       arity _                                          = error "MangleTcToCC.superfeature.arity: Impossible error!\n"

       transformEq (TcEquat [] e) = TcEquat [TcPatternAppl (TcUnqualName "_Nil") []] e
       transformEq _		  = error "MangleTcToCC.superfeature.transformEq: Impossible error!\n"

       transformDecl fdecl@(TcLocalFuncDecl t a n e) 
		     | (n == (TcName "main")) || ((arity fdecl) /= 0) = fdecl
		     | otherwise				      = TcLocalFuncDecl t a n (map transformEq e)
       transformDecl patterndecl				      = patterndecl

       constants = [n | cdef@(TcLocalFuncDecl _ _ (TcName n) _) <- localdefs, 
			arity cdef == 0, 
			n /= "main"]

       subst = listToFM [(n', TcExprAppl n' (TcExprConstr (TcUnqualName "_Nil"))) | n <- constants, 
										    let n' = nameToExpr n]
       nameToExpr n = TcExprVar (TcUnqualName n)
       namesToExprs = map nameToExpr

       substituteLDefs s                              = map (substituteLDef s)
       substituteLDef  s (TcLocalFuncDecl t a n es)   = TcLocalFuncDecl t a n (substituteEqus s es)
       substituteLDef  s (TcLocalPatternDecl p expr)  = let s' = delListFromFM s $ namesToExprs (freePattern p)
							in  TcLocalPatternDecl p (substituteExpr s' expr)
       substituteEqus  s			      = map (substituteEqu s)
       substituteEqu   s (TcEquat ps expr)	      = let s' = delListFromFM s $ namesToExprs (freePatterns ps)
							in  TcEquat ps (substituteExpr s' expr)
       substituteExpr  s (TcExprLambda ps expr)	      = let s' = delListFromFM s $ namesToExprs (freePatterns ps)
							in  TcExprLambda ps (substituteExpr s' expr)
       substituteExpr  s (TcExprFreeDecl vs expr)     = let s' = delListFromFM s (namesToExprs [v | TcName v <- vs])
							in  TcExprFreeDecl vs (substituteExpr s' expr)
       substituteExpr  s (TcExprLet ldef expr)	      = let s' = delListFromFM s $ namesToExprs (definedByLDef ldef)
							in  TcExprLet (substituteLDef s' ldef) (substituteExpr s' expr)
       substituteExpr  s (TcExprLetrec ldefs expr)    = let s' = delListFromFM s $ namesToExprs (definedByLDefs ldefs)
							in  TcExprLetrec (substituteLDefs s' ldefs) (substituteExpr s' expr)
       substituteExpr  s (TcExprAppl expr1 expr2)     = TcExprAppl (substituteExpr s expr1) (substituteExpr s expr2)
       substituteExpr  s v@(TcExprVar _)	      = lookupWithDefaultFM s v v
       substituteExpr  _ e			      = e
   in  substituteLDefs subst (map transformDecl localdefs)   



--                 result  search variables
--                    |         |
type SearchVarTab = (Int, [(Int, String)])


mangleTcToCC :: TcModule -> MaybeError (CCProgram, SearchVarTab)
----------------------------------------------------------------
mangleTcToCC (TcModule matchStrat dd ld) = 
   let suppliedWith f supp = fst ((unSupplyM f) supp)
   in  do ctab <- collectDataDecls dd
	  mangleInit ld `suppliedWith` (initSupplier ctab matchStrat)





mangleInit :: [TcLocalDef] -> SupplyM (CCExpr, SearchVarTab)
------------------------------------------------------------
mangleInit localdefs =
   let mainBody		   = [equations | TcLocalFuncDecl _ _ (TcName "main") equations <- localdefs]
       lambdas    []	 e = e								-- todo : fold l/r?
       lambdas    (v:vs) e = CCExprLambda v (lambdas vs e)
       frees	  []	 e = e								-- todo : fold l/r?
       frees	  (v:vs) e = CCExprFreeDecl v (frees vs e)
       constrFunctions = let fun ((CCName n), arity, constr) =
			        do fc   <- newVar
				   vars <- localNewVars arity
				   newBinding (TcName n) fc
				   return $ CCFunDef fc (lambdas vars (CCExprConstr constr vars))
			 in  do allConstructors <- allDefinedConstructors
			 	mapM fun [(n,a,c) | c@(CCDataConstr n _ a) <- allConstructors , a > 0]
   in  if null mainBody then
	  fail $ "main undefined!"
       else 
          let (equ:r) = head mainBody
	  in  if not (null r) then
		 fail $ "main multiply defined!"
	      else
   	         let searchVars   = freeEqu equ \\ definedByLDefs localdefs
		     nrSearchVars = length searchVars
		 in  do svars   <- newVars nrSearchVars
			mapM (\ (bind, to) -> newBinding (TcName bind) to) (zip searchVars svars)
			let localdefs' = superfeature localdefs
			instantiateDefinedVars localdefs'
			cfuns   <- constrFunctions
			fundefs <- mangleLocalDefs localdefs'
			mainVar@(CCVariable mainId) <- lookupBinding (TcName "main")
			let expr = frees svars $ CCExprLet (fundefs ++ cfuns) (CCExprStrict (CCExprVar mainVar))
			return (expr, (mainId, zip (map (\ (CCVariable id) -> id) svars) searchVars))
			


mangleExpr :: TcExpr -> SupplyM CCExpr
--------------------------------------
-- lambda
mangleExpr (TcExprLambda patterns expr) =
   tDTree (Just defaultAnnot) [TcEquat patterns expr]
--                     |
--                   HACK!! 
-- should depend on the type of expr, which we don't know

-- free
mangleExpr (TcExprFreeDecl [] expr) =
   mangleExpr expr
mangleExpr (TcExprFreeDecl (x:xs) expr) =
   do x'     <- newVar
      newBinding x x'
      expr'  <- mangleExpr (TcExprFreeDecl xs expr)
      return $ CCExprFreeDecl x' expr'
-- let
mangleExpr (TcExprLet ldef expr) =
   mangleExpr (TcExprLetrec [ldef] expr)
-- letrec
mangleExpr (TcExprLetrec ldefs expr) = 
   do instantiateDefinedVars ldefs
      ldefs' <- mangleLocalDefs ldefs
      expr'  <- mangleExpr	expr
      return $ CCExprLet ldefs' expr'
-- app
-- builtins (& =:= |)
-- this is brutal force :-(
-- &
mangleExpr (TcExprAppl (TcExprAppl (TcExprVar (TcUnqualName "&")) expr1) expr2) =
   do expr1' <- shadowing (mangleExpr expr1)
      expr2' <- mangleExpr expr2
      return $ CCExprAnd expr1' expr2'
-- =:=
mangleExpr (TcExprAppl (TcExprAppl (TcExprVar (TcUnqualName "=:=")) expr1) expr2) =
   do expr1' <- shadowing (mangleExpr expr1)
      expr2' <- mangleExpr expr2
      return $ CCExprUnify expr1' expr2'
-- |
-- remember that (as we have no type information) we assume guards to be
-- constraints always!
mangleExpr (TcExprAppl (TcExprVar (TcUnqualName "_guard")) guardedExprs) =
   mangleGuardedExprs guardedExprs
-- other applications
-- if a variable is applied already, no additional indirection is necessary
mangleExpr (TcExprAppl expr (TcExprVar v)) =
   do v'    <- mangleKnownVar  v
      expr' <- mangleExpr expr
      return $ CCExprAppl expr' v'
mangleExpr (TcExprAppl expr1 expr2) =
   do v      <- newVar
      expr1' <- shadowing (mangleExpr expr1)
      expr2' <- mangleExpr expr2
      return $ CCExprLet [CCFunDef v expr2'] (CCExprAppl expr1' v)
mangleExpr (TcExprVar var) = 
   do var'   <- mangleKnownVar var
      return $ CCExprVar var'
mangleExpr (TcExprConstr c) = 
   do const <- mangleMaybeConstant c
      case const of
	   Nothing -> do fc <- mangleConstr c
			 return $ CCExprVar fc
	   Just c' -> return $ CCExprConstr c' []
mangleExpr (TcExprLiteral l) = 
   do l' <- mangleLiteral l
      return $ CCExprConstr l' []






instantiateDefinedVars :: [TcLocalDef] -> SupplyM ()
----------------------------------------------------
instantiateDefinedVars defs =
   mapM_ (\ n -> do v <- newVar
		    newBinding (TcName n) v)
	 (definedByLDefs defs)


mangleLocalDefs :: [TcLocalDef] -> SupplyM [CCFunDef]
-----------------------------------------------------
mangleLocalDefs defs = 
   let mangleLocalDef (TcLocalFuncDecl _ annot v equs) =
          do v'	   <- lookupBinding v
	     expr  <- shadowing (tDTree annot equs)
	     return $ [CCFunDef v' expr]
       mangleLocalDef pdecl =
          shadowing (tPDecl pdecl)
   in concatMapM mangleLocalDef defs
		

mangleGuardedExprs :: TcExpr -> SupplyM CCExpr
----------------------------------------------
mangleGuardedExprs guardedExprs =
   let fromList (TcExprAppl (TcExprAppl (TcExprConstr (TcUnqualName "_Cons")) headExpr) (TcExprConstr (TcUnqualName "_Nil"))) =
          fromPair headExpr
       fromList (TcExprAppl (TcExprAppl (TcExprConstr (TcUnqualName "_Cons")) headExpr) tailExpr) =
          do expr1 <- shadowing (fromPair headExpr)
	     expr2 <- fromList tailExpr
	     return $ CCExprOr expr1 expr2
       fromList _ =
          error "MangleTcToCC.mangleGuardedExprs: unexpected form of guarded expressions!\n"
       fromPair (TcExprAppl (TcExprAppl (TcExprConstr (TcUnqualName "_Tuple2")) fstExpr) sndExpr) =
	  do fstExpr' <- shadowing (mangleExpr fstExpr)
	     sndExpr' <- mangleExpr sndExpr
	     return $ CCExprGuard fstExpr' sndExpr'
       fromPair _ =
          error "MangleTcToCC.mangleGuardedExprs: unexpected form of guarded expressions!\n"
   in  fromList guardedExprs


mangleKnownVar :: TcQVariableID -> SupplyM CCVariable
-----------------------------------------------------
mangleKnownVar (TcUnqualName n) = 
   lookupBinding (TcName n)
mangleKnownVar (TcQualName _ n) = 
   fail $ "Qualified variable names not (yet) supported: " ++ n


mangleMaybeConstant :: TcQDataConstrID -> SupplyM (Maybe CCDataConstr)
----------------------------------------------------------------------
mangleMaybeConstant n =
   do c@(CCDataConstr _ _ arity) <- lookupCTab n
      if arity == 0 then
	return $ Just c
       else
	return Nothing


mangleConstr :: TcQDataConstrID -> SupplyM CCVariable
----------------------------------------------------
mangleConstr (TcUnqualName name) =
   lookupBinding (TcName name)
mangleConstr (TcQualName _ n) =
   fail $ "Qualified constructor names not (yet) supported: " ++ n
mangleLiteral (LiteralInt int) = 
   return $ CCDataConstr (CCName (show int)) int 0





-- transformation functions from the thesis --------------------------
-- i.e. tDTree and tPDecl                   --------------------------
----------------------------------------------------------------------



-- tDTree --
------------

-- helpers: 

 
-- minPosition defines the matching strategy

isCBased :: TermPosition -> (Int, Substitution, TcEquat) -> SupplyM Bool
------------------------------------------------------------------------
isCBased o equ = do t <- termAtPositionTab o equ
		    return $ case t of
				  TcPatternAppl _ _  -> True
				  TcPatternLiteral _ -> True
				  _		     -> False

minPosition :: MatchStrategy -> [(Int, Substitution, TcEquat)] -> [TermPosition] -> SupplyM (TermPosition, Int)
---------------------------------------------------------------------------------------------------------------
minPosition strat d os 
   | strat == OptMatch = maxCBased os
   | strat == FunctionalMatch = let o = minimum os -- this is really nice :-)
				in  do nr <- nrCBased d o
				       return (o,nr)
   | otherwise = error "MangleTcToCC.minPosition: Impossible error!\n" -- Don't know why a warning occurs without this definition...
   where nrCBased equs o = do boollist <- mapM (isCBased o) equs
			      return $ length (filter id boollist)
	 maxCBased' _      maxpos []     = return maxpos
	 maxCBased' maxval maxpos (o:os) = do nr <- nrCBased d o
					      if nr > maxval then maxCBased' nr    (o,nr)  os
							     else maxCBased' maxval maxpos os
	 maxCBased (o:os) = do nr <- nrCBased d o
			       maxCBased' nr (o,nr) os
         maxCBased []     = error "MangleTcToCC.minPosition.maxCBased: Impossible error!\n"







-- tDTree itself

type Tilde = FiniteMap TermPosition CCVariable

tDTree :: Maybe Annotation -> [TcEquat] -> SupplyM CCExpr
------------------------------------------------------------------------
tDTree annot equs =
   let annotation               = case annot of
				       Just a  -> a
				       Nothing -> defaultAnnot -- should not happen! (HACK)
       n                        = length $ head [patterns | (TcEquat patterns _) <- equs] -- the functions arity
       positions		= [[p] | p <- [0..n-1]]
       lambdas 0 tilde          = do resetPositionTab
				     subst <- copySubst
				     -- because of possible variable name clashes in the definitions
				     -- we provide each defining equation with it's own substitution
				     tDTree' (zip3 [1..] (repeat subst) equs) positions tilde annotation
       lambdas m tilde		= do v    <- newVar
				     expr <- lambdas (m-1) (addToFM tilde [abs (m - n)] v)
				     return $ CCExprLambda v expr
   in  lambdas n emptyFM


tDTree' :: [(Int,Substitution,TcEquat)] -> [TermPosition] -> Tilde -> Annotation -> SupplyM CCExpr
--------------------------------------------------------------------------------------------------
tDTree' d [] _ _ =
   -- all argument positions handled
   let ors [(subst,e)]    = do resetSubst subst
			       mangleExpr e
       ors ((subst,e):es) = do resetSubst subst
			       liftM2 CCExprOr (shadowing (mangleExpr e)) (ors es)
       ors []             = error "MangleTcToCC.tDTree'.ors: Impossible error!\n"
   in  ors [(subst,expr) | (_,subst,(TcEquat _ expr)) <- d]
tDTree' d o tilde a =
   -- there are still unhandled argument positions
   do mstrat <- matchStrategy
      (minO, nrCBased) <- minPosition mstrat d o
      if nrCBased == 0
	 then -- at position minO all arguments are a (evtl. anonymous) variable
	      do vars <- mapM (termAtPositionTab minO) d
   		 let newVar = fromJust $ lookupFM tilde minO
		     d'	    = map (\ (e@(id,subst,equ),pattern) ->
				     case pattern of
					  TcPatternVar argVar -> (id,newBindingSubst subst argVar newVar,equ)
					  TcPatternAnonymous  -> e
                                          _                   -> error "MangleTcToCC.tDTree': Impossible error!\n")
			      (zip d vars)
		 tDTree' d' (delete minO o) tilde a
	 else if nrCBased == length d
		 then -- at position minO all arguments are constructor based
		      do alts <- alternatives d o tilde a minO
			 return $ CCExprCase a (fromJust $ lookupFM tilde minO) alts
		 else -- at position minO are cbased terms as well as variables
		      -- => divide and conquer
		      do (d',d'') <- partitionM (isCBased minO) d
			 -- no explicit subst-copying necessary, 
			 -- as each element of d has it's own subst
			 expr1    <- tDTree' d'  o tilde a
			 expr2    <- tDTree' d'' o tilde a
			 return $ CCExprOr expr1 expr2




alternatives ::  [(Int,Substitution,TcEquat)] -> [TermPosition] -> Tilde -> Annotation -> TermPosition -> SupplyM [CCAlternative]
---------------------------------------------------------------------------------------------------------------------------------
alternatives d o tilde a minO = 
   do cBasedTerms <- mapM (termAtPositionTab minO) d
      let alternative constructors unhandled c@(CCDataConstr _ _ arity) 
	    | c `elem` unhandled =
	      shadowing $
              do vars <- newVars arity
	         return $ CCAlternative c vars CCExprFail
	    | otherwise = 
	      let d_i	= [d' | (c',d') <- zip constructors d, c == c' ]
		  newPos  = [minO ++ [j] | j <- [0..(arity-1)]]
		  o_i     = (delete minO o) ++ newPos
	      in  shadowing $
		  do vars <- newVars arity
		     let tilde_i = foldr (\ (pos,var) tilde' -> addToFM tilde' pos var) tilde (zip newPos vars)
		     expr <- tDTree' d_i o_i tilde_i a
		     return $ CCAlternative c vars expr
	  isLiteral = case head cBasedTerms of
			   (TcPatternLiteral _) -> True
			   _			-> False
      if isLiteral 
	 then let constructors = [l' | TcPatternLiteral l <- cBasedTerms, l' <- mangleLiteral l]
	      in  mapM (alternative constructors []) (nub constructors)
	 else do let (TcPatternAppl example _) = head cBasedTerms
		 constructors    <- mapM lookupCTab [name | TcPatternAppl name _ <- cBasedTerms]
		 allConstructors <- lookupCTabConstrsOfType example
		 let unhandledConstructors = allConstructors \\ constructors
		 mapM (alternative constructors unhandledConstructors) allConstructors






-- tPDecl --
------------

-- unfortunately, tPDecl is not tested, as
-- MangleRdrToTc doesn't support pattern declarations by now :-(


tPDecl :: TcLocalDef -> SupplyM [CCFunDef]
------------------------------------------
tPDecl (TcLocalPatternDecl pattern@(TcPatternAppl _ _) expr) =
   let unwindPDecl (TcPatternVar var, expr) = 
          do var'  <- lookupBinding var     -- the variable is known already!
	     return $ [CCFunDef var' expr]
       unwindPDecl (TcPatternAppl c patterns, expr) =
          do ccConstr@(CCDataConstr _ _ arity) <- lookupCTab              c
	     allccConstrs		       <- lookupCTabConstrsOfType c
	     let unhandledccConstrs = delete ccConstr allccConstrs
	     errors <- mapM (\ c'@(CCDataConstr _ _ a) -> do vars  <- localNewVars a
							     return $ CCAlternative c' vars (CCExprError CCErrorPatternDecl))
			    unhandledccConstrs
	     alts   <- do vars  <- localNewVars arity
			  return $ map (\ v -> CCAlternative ccConstr vars (CCExprVar v)) vars
             var    <- newVar
	     let cases = [CCExprCase defaultAnnot var alts' | alts' <- [(a:errors) | a <- alts]]
--				       |
--				     HACK!
	     proceed <- concatMapM unwindPDecl (zip patterns cases)
	     return $ (CCFunDef var expr):proceed
       unwindPDecl _ = 
          return [] -- either anonymous variables as literals force no further computation...
   in do expr' <- mangleExpr expr
	 unwindPDecl (pattern, expr')
tPDecl (TcLocalPatternDecl (TcPatternAnonymous) _) =
   return [] -- this is really unexpected ;-)
tPDecl (TcLocalPatternDecl (TcPatternLiteral _) _) =
   return [] -- this also
tPDecl _ =
   error "tPDecl: Impossible!\n"






-- Variable ID Supplier Monad ----------------------------------------
----------------------------------------------------------------------

-- 1 st element of the supplier monad
-- Substitution
-------------------------------------


newtype Substitution = Substitution (FiniteMap String CCVariable)

emptySubst = Substitution emptyFM

newBindingSubst :: Substitution -> TcName -> CCVariable -> Substitution
------------------------------------------------------------------
newBindingSubst (Substitution subst) (TcName bindName) to = 
   Substitution $ addToFM subst bindName to

lookupBindingSubst :: Substitution -> TcName -> Maybe CCVariable
-----------------------------------------------------------
lookupBindingSubst (Substitution subst) (TcName name) =
   lookupFM subst name





-- 2nd element of the supplier monad
-- VSupp
------------------------------------


newtype VSupp = VSupp Int

initVSupp = VSupp 0


newVarVSupp :: VSupp -> (VSupp, CCVariable)
----------------------------------------
newVarVSupp (VSupp next) = 
   (VSupp (next + 1), CCVariable next)


newVarsVSupp :: VSupp -> Int -> (VSupp, [CCVariable])
----------------------------------------------------
newVarsVSupp (VSupp next) num =
   (VSupp (next + num), map CCVariable [next..(next+num-1)])






-- 3rd element of the supplier monad
-- PTable

-- Position

type TermPosition = [Int]

-- as we don't want to calculate the terms at the positions again and again,
-- we memorize already calculated results

type PTable = FiniteMap (Int, TermPosition) TcPattern

emptyPTable = emptyFM


termAtPositionEquat :: TermPosition -> TcEquat -> TcPattern
-----------------------------------------------------------
termAtPositionEquat (p:ps) (TcEquat patterns _) =
   termAtPosition ps (patterns!!p)
termAtPositionEquat [] _ = error "MangleTcToCC.termAtPositionEquat: Impossible error!\n"

termAtPosition :: TermPosition -> TcPattern -> TcPattern
--------------------------------------------------------
termAtPosition [] t = t
termAtPosition (p:ps) (TcPatternAppl _ patterns) = termAtPosition ps (patterns!!p)
termAtPosition _ _ = error "MangleTcToCC.termAtPosition: Position doesn't exist!\n"


-- 4th of the supplier monad
-- ConstrTable

type ConstrTable = FiniteMap TcName (CCDataConstr, [CCDataConstr])

newtype Arity = Arity Int






-- SupplyM --
-------------

newtype SupplyM a = SupplyM (Supplier -> 
			     (MaybeError a, Supplier))


data Supplier = Supplier {
			   subst  :: Substitution
			 , vsupp  :: VSupp
			 , ptab	  :: PTable
			 , ctab	  :: ConstrTable
			 , mstrat :: MatchStrategy
			 }



initSupplier ctab' mstrat' = Supplier { subst   = emptySubst
				      , vsupp   = initVSupp
				      , ptab    = emptyPTable
				      , ctab    = ctab'
				      , mstrat  = mstrat' }


instance Monad SupplyM where
   m >>= k    = SupplyM (\ supp -> 
			   let (val, supp') = (unSupplyM m) supp
			   in  case val of
				    Error e        -> (Error e,supp')
				    NullProblemo a -> unSupplyM (k a) supp')
   return val = SupplyM (\ supp -> (return val, supp))

unSupplyM (SupplyM m)      = m



shadowing f = do supp <- copySupplier
		 res <- f
		 resetSupplier supp
		 return res


copySupplier        = SupplyM (\ supp                 -> (NullProblemo supp,         supp)) 

resetSupplier supp  = SupplyM (\ _                    -> (NullProblemo (),           supp))

copySubst	    = SupplyM (\ supp		      -> (NullProblemo (subst supp), supp)) 

resetSubst subst'   = SupplyM (\ supp		      -> (NullProblemo (),	     supp{ subst = subst' }))


resetPositionTab    = SupplyM (\ supp		      -> (NullProblemo (),	     supp{ ptab = emptyPTable }))


newVar		    = SupplyM (\ supp -> 
				 let (vsupp',var) = newVarVSupp (vsupp supp)
				 in  (NullProblemo var, supp{ vsupp = vsupp' }))

newVars n	    = SupplyM (\ supp -> 
				 let (vsupp',vars) = newVarsVSupp (vsupp supp) n
				 in  (NullProblemo vars, supp{ vsupp = vsupp' }))

localNewVars n	    = SupplyM (\ supp -> 
				 let (_,vars) = newVarsVSupp (vsupp supp) n
				 in  (NullProblemo vars, supp))


newBinding bind to  = SupplyM (\ supp -> 
				 let subst' = newBindingSubst (subst supp) bind to
				 in  (NullProblemo (), supp{ subst = subst' }))


lookupBinding v@(TcName n) = 
   SupplyM (\ supp -> 
	      case lookupBindingSubst (subst supp) v of
		   Nothing -> (Error $ "Undefined variable: " ++ n, supp)
		   Just v' -> (NullProblemo v', supp))

lookupMaybeBinding v = 
   SupplyM (\ supp -> 
	      (NullProblemo (lookupBindingSubst (subst supp) v), supp))


termAtPositionTab o (id,_,equ) = 
   SupplyM (\ supp -> 
	      case lookupFM (ptab supp) (id,o) of
		   Nothing -> let t = termAtPositionEquat o equ
			      in  (NullProblemo t, supp{ ptab = addToFM (ptab supp) (id,o) t })
		   Just t  -> (NullProblemo t, supp))


lookupCTab (TcUnqualName n) =
   SupplyM (\ supp -> 
	      case lookupFM (ctab supp) (TcName n) of
		   Nothing -> (Error $ "Undefined constructor: " ++ n, supp)
		   Just e  -> (NullProblemo $ fst e, supp))


lookupCTab (TcQualName _ n) =
   SupplyM (\ supp -> 
	      (Error $ "Qualified constructor names not (yet) supported: " ++ n , supp))


lookupCTabConstrsOfType (TcUnqualName n) =
   SupplyM (\ supp -> 
	      case lookupFM (ctab supp) (TcName n) of
		   Nothing -> (Error $ "Undefined constructor: " ++ n, supp)
		   Just e  -> (NullProblemo $ snd e, supp))


lookupCTabConstrsOfType (TcQualName _ n) =
   SupplyM (\ supp -> 
	      (Error $ "Qualified constructor names not (yet) supported: " ++ n , supp))


allDefinedConstructors =
   SupplyM (\ supp -> 
	   (NullProblemo $ map fst (eltsFM (ctab supp)), supp))


matchStrategy =
   SupplyM (\ supp -> 
	   (NullProblemo $ mstrat supp, supp))




-- SUP's work in MangleTcToCur: --------------------------------------
----------------------------------------------------------------------
-- Collect information about data type declarations

-- For the pattern matching compiler, all constructors belonging to the same
-- type as a given constructor are needed, too.

-- If the type checker test for repeated definitions, the check below can
-- be removed.
collectDataDecls :: [TcDataDeclaration] -> MaybeError ConstrTable
collectDataDecls dataDecls =
   foldM (\table (name, entry) ->
             case lookupFM table name of
                Nothing -> return (addToFM table name entry)
                Just _  -> fail . render $ text "Repeated definition for constructor function \"" <+>
                                           pretty undefined name <+> text "\"")
         emptyFM
         [ (tn, (cInfo, map snd cInfos)) |
               TcDataDeclaration _ _ constrDecls <- preludeDataDecls ++ dataDecls,
               let cInfos = [ (TcName c,
                               CCDataConstr (CCName c) t (length args)) |
                              (t, TcConstrDecl (TcName c) args) <- zip [0..] constrDecls ],
               (tn, cInfo) <- cInfos ]

-- SUP: HACK HACK HACK
preludeDataDecls :: [TcDataDeclaration]
preludeDataDecls = 
   [ TcDataDeclaration (TcName "_List") [TcName "a"]
                       [TcConstrDecl (TcName "_Nil")  [],
                        TcConstrDecl (TcName "_Cons") [TcTypeVar (TcName "a"),
                                                       TcTypeConstr (TcUnqualName "_List") [TcTypeVar (TcName "a")]]]]





