module ParseMonad where

-- import IOExts(trace)

import CommonSyntax(SrcLoc(..))

data ParseResult a = Ok ParseState a | Failed String
	deriving Show

data LexContext = NoLayout | Layout Int
	deriving (Eq,Ord,Show)

type ParseState = [LexContext]

type P a
     =  String			-- input string
     -> SrcLoc			-- location of last token read
     -> Int				-- current column
     -> ParseState			-- layout info.
     -> ParseResult a

thenP :: P a -> (a -> P b) -> P b
m `thenP` k = \i l c s -> 
		case m i l c s of 
		    Failed s' -> Failed s'
		    Ok s' a -> case k a of k' -> k' i l c s'

thenP_ :: P a -> P b -> P b
m `thenP_` k = m `thenP` \_ -> k

mapP :: (a -> P b) -> [a] -> P [b]
mapP _ [] = returnP []
mapP f (a:as) = 
     f a `thenP` \b ->
     mapP f as `thenP` \bs ->
     returnP (b:bs)

returnP :: a -> P a
returnP a = \_ _ _ s -> Ok s a

failP :: String -> P a
failP err = \_ _ _ _ -> Failed err

getSrcLoc :: P SrcLoc
getSrcLoc = \_ l _ s -> Ok s l

getContext :: P [LexContext]
getContext = \_ _ _ s -> Ok s s

pushContext :: LexContext -> P ()
pushContext ctxt = 
	--trace ("pushing lexical scope: " ++ show ctxt ++"\n") $
	\_ _ _ s -> Ok (ctxt:s) ()

popContext :: P ()
popContext = \_ _ _ stk ->
      case stk of
	   (_:s) -> --trace ("popping lexical scope, context now "++show s ++ "\n") $ 
	            Ok s ()
	   []    -> error "Internal error: empty context in popContext"

parseError :: String -> P a
parseError s = \r (SrcLoc y x) -> 
    failP (show y ++ ":" ++ show x ++ ": " ++ s) r (SrcLoc y x)
