{- Hey Emacs, this is -*- haskell -*-

Changes to grammar in Curry report:

   * module header is optional
   * "pragma choice" is allowed
   * infix constructors added
   * Natural accepts all INT_LITERALs (checked in monad)
   * TypeExpr makes binding power clear
   * InfixConstrID and QInfixConstrID (not yet in lexer) introduced
   * Only Int literals for now
-}

> {
> module Parser(parseModule, parseExpression) where

> import Char(isDigit)
> import Lexer
> import CommonSyntax
> import ReaderSyntax
> import ErrorMonad
> import ParseMonad
> }
> 
> %token -- reserved words
>        'choice'           { TokenChoice               }
>        'data'             { TokenData                 }
>        'else'             { TokenElse                 }
>        'eval'             { TokenEval                 }
>        'flex'             { TokenFlex                 }
>        'free'             { TokenFree                 }
>        'if'               { TokenIf                   }
>        'import'           { TokenImport               }
>        'in'               { TokenIn                   }
>        'infix'            { TokenInfix                }
>        'infixl'           { TokenInfixl               }
>        'infixr'           { TokenInfixr               }
>        'let'              { TokenLet                  }
>        'module'           { TokenModule               }
>        'optmatch'         { TokenOptmatch             }
>        'pragma'           { TokenPragma               }
>        'renaming'         { TokenRenaming             }
>        'rigid'            { TokenRigid                }
>        'then'             { TokenThen                 }
>        'to'               { TokenTo                   }
>        'where'            { TokenWhere                }
>        -- literals
>        INT_LITERAL        { TokenIntLiteral $$        }
>        -- names
>        IDENTIFIER         { TokenIdentifier        $$ }
>        DOTTED_IDENTIFIER  { TokenDottedIdentifier  $$ }
>        CONSTRUCTOR        { TokenConstructor       $$ }
>        DOTTED_CONSTRUCTOR { TokenDottedConstructor $$ }
>        INFIXCONSTRUCTOR   { TokenInfixConstructor  $$ }
>        -- operators/separators
>        '->'               { TokenArrow                }
>        '='                { TokenEquals               }
>        '|'                { TokenBar                  }
>        '::'               { TokenSig                  }
>        '\\'               { TokenLambda               }
>        '..'               { TokenDotDot               }
>        ';'                { TokenSemicolon            }
>        ','                { TokenComma                }
>        '_'                { TokenAnonymous            }
>        '('                { TokenLParen               }
>        ')'                { TokenRParen               }
>        '['                { TokenLBracket             }
>        ']'                { TokenRBracket             }
>        '{'                { TokenLBrace               }
>        '}'                { TokenRBrace               }
>        VIRTUALBRACE       { TokenVRBrace              }
>        -- infix operators
>        '-'                { TokenMinus                }
>        INFIXOPID          { TokenInfixOpId $$         }
>        -- not used yet
>        EXTERNAL           { TokenExternal             }
>        SCANNER_ERROR      { TokenError $$             }

> %monad { P } { thenP } { returnP }
> %lexer { lexer } { EOF }
> %name parseCurry
> %tokentype { Token }
> %%

-- SUP: MEGA HACK for multiple parser entries!!!

> Start :: { ParseType }
>    : '|' Module                                                       { ItsAModule $2      }
>    | ';' Expr                                                         { ItsAnExpression $2 }

---------- module structure ----------------------------------------------------------------------

> Module :: { RdrModule }
>    : 'module' ModuleID MaybeExports 'where' Block                     { RdrModule (Just ($2, $3)) $5 }
>    |                                        Block                     { RdrModule Nothing         $1 }

> ModuleID :: { RdrModuleID }
>    : IDENTIFIER                                                       { RdrModuleID [$1] }
>    | DOTTED_IDENTIFIER                                                { RdrModuleID  $1  }
>    | CONSTRUCTOR                                                      { RdrModuleID [$1] }
>    | DOTTED_CONSTRUCTOR                                               { RdrModuleID  $1  }

> MaybeExports :: { Maybe [RdrExport] }
>    : {- EMPTY -}                                                      { Nothing }
>    | '(' MaybeExportList ')'                                          { Just $2 }

> MaybeExportList :: { [RdrExport] }
>    : {- EMPTY -}                                                      { []         }
>    | ExportList                                                       { reverse $1 }

> ExportList :: { [RdrExport] }
>    :                Export                                            { [$1]    }
>    | ExportList ',' Export                                            { $3 : $1 }

> Export :: { RdrExport }
>    : FunctionName                                                     { RdrExportFunc $1         }
>    | TypeConstrID                                                     { RdrExportTypeAbstract $1 }
>    | TypeConstrID '(' MaybeDataConstrIDs ')'                          { RdrExportTypeWith $1 $3  }
>    | TypeConstrID '(' '..' ')'                                        { RdrExportTypeAll $1      }
>    | 'module' ModuleID MaybeFuncOrTypeList                            { RdrExportModule $2 $3    }

> MaybeDataConstrIDs :: { [RdrDataConstrID] }
>    : {- EMPTY -}                                                      { []         }
>    | DataConstrIDs                                                    { reverse $1 }

> DataConstrIDs :: { [RdrDataConstrID] }
>    :                   DataConstrID                                   { [$1]    }
>    | DataConstrIDs ',' DataConstrID                                   { $3 : $1 }

> MaybeFuncOrTypeList :: { Maybe [RdrImportEntity] }
>    : {- EMPTY -}                                                      { Nothing }
>    | '(' MaybeFuncOrTypes ')'                                         { Just $2 }

> MaybeFuncOrTypes :: { [RdrImportEntity] }
>    : {- EMPTY -}                                                      { []         }
>    | FuncOrTypes                                                      { reverse $1 }

> FuncOrTypes :: { [RdrImportEntity] }
>    :                 FuncOrType                                       { [$1]    }
>    | FuncOrTypes ',' FuncOrType                                       { $3 : $1 }

> FuncOrType :: { RdrImportEntity }
>    : FunctionName                                                     { RdrImportFunc $1 }
>    | TypeConstrID                                                     { RdrImportType $1 }

> Block :: { ([RdrImport], [RdrPragma], [RdrFixityDecl], [RdrBlockDecl]) }
>    : '{' LayoutOff BlockElems '}'                                     { $3 }
>    |     LayoutOn  BlockElems Close                                   { $2 }

> BlockElems :: { ([RdrImport], [RdrPragma], [RdrFixityDecl], [RdrBlockDecl]) }
>    : Imports Pragmas FixityDeclarations MaybeBlockDeclarations        { (reverse $1, reverse $2, reverse $3, $4) }

> Imports :: { [RdrImport] }
>    : {- EMPTY -}                                                      { []      }
>    | Imports Import ';'                                               { $2 : $1 }

> Pragmas :: { [RdrPragma] }
>    : {- EMPTY -}                                                      { []      }
>    | Pragmas Pragma ';'                                               { $2 : $1 }

> FixityDeclarations :: { [RdrFixityDecl] }
>    : {- EMPTY -}                                                      { []      }
>    | FixityDeclarations FixityDeclaration ';'                         { $2 : $1 }

> Import :: { RdrImport }
>    : 'import' ModuleID MaybeFuncOrTypeList Renamings                  { RdrImport $2 $3 (reverse $4) }

> Renamings :: { [RdrRenaming] }
>    : {- EMPTY -}                                                      { []      }
>    | Renamings Renaming                                               { $2 : $1 }

> Renaming :: { RdrRenaming }
>    : 'renaming' FunctionName 'to' FunctionName                        { RdrRenaming $2 $4 }
>    | 'renaming' TypeConstrID 'to' TypeConstrID                        { RdrRenaming $2 $4 }

> Pragma :: { RdrPragma }
>    : 'pragma' PragmaOption                                            { $2 }

> PragmaOption :: { RdrPragma }
>    : Annotation                                                       { RdrPragmaAnnot $1 }
>    | MatchStrategy                                                    { RdrPragmaMatch $1 }

> MatchStrategy :: { MatchStrategy }
>    : 'optmatch'                                                       { OptMatch }

> MaybeBlockDeclarations :: { [RdrBlockDecl] }
>    : {- EMPTY -}                                                      { []         }
>    | BlockDeclarations                                                { reverse $1 }

> BlockDeclarations :: { [RdrBlockDecl] }
>    :                       BlockDeclaration                           { [$1]    }
>    | BlockDeclarations ';' BlockDeclaration                           { $3 : $1 }
 
> BlockDeclaration :: { RdrBlockDecl }
>    : DataDeclaration                                                  { RdrDataDecl $1 }
>    | FunctionDeclaration                                              { RdrFuncDecl $1 }

---------- declaration of a new algebraic type ---------------------------------------------------

> DataDeclaration :: { RdrDataDeclaration }
>    : 'data' TypeDeclaration                                           { $2 }

> TypeDeclaration :: { RdrDataDeclaration }
>    : TypeConstrID MaybeTypeVarIDs '=' ConstrDeclarations              { RdrDataDeclaration $1 (reverse $2) (reverse $4) }

> TypeConstrID :: { RdrTypeConstrID }
>    : CONSTRUCTOR                                                      { RdrName $1 }

> MaybeTypeVarIDs :: { [RdrTypeVarID] }
>    : {- EMPTY -}                                                      { []      }
>    | MaybeTypeVarIDs TypeVarID                                        { $2 : $1 }

> ConstrDeclarations :: { [RdrConstrDecl] }
>    :                        ConstrDeclaration                         { [$1]    }
>    | ConstrDeclarations '|' ConstrDeclaration                         { $3 : $1 }
 
-- SUP: We hackily parse a constructor declaration as a BasicType and split the resulting
--      type applications into constructor/argument parts later. This avoids 3 r/r conflicts
--      for something like "data Foo = Baz a b c :+: Blah", too. It is not decidable if Baz
--      is a type constructor or a data constructor with a finite lookahead.

> ConstrDeclaration :: { RdrConstrDecl }
>    : BasicType                                                        {% splitBasicType $1        }
>    | BasicType InfixConstrID BasicType                                { RdrConstrDecl $2 [$1, $3] }

> DataConstrID :: { RdrDataConstrID }
>    : CONSTRUCTOR                                                      { RdrName $1 }

> InfixConstrID :: { RdrName }
>    : INFIXCONSTRUCTOR                                                 { RdrName $1 }

---------- type ----------------------------------------------------------------------------------

> TypeExpr :: { RdrType }      -- right recursive (no 'reverse' on TypeFuns yet)
>    : BasicType                                                        { $1                }
>    | BasicType '->' TypeExpr                                          { RdrTypeFunc $1 $3 }

> BasicType :: { RdrType }
>    : AtomicType                                                       { $1                                                 }
>    | QTypeConstrID AtomicTypes                                        { foldl1 RdrTypeAppl (RdrTypeConstr $1 : reverse $2) }

> AtomicTypes :: { [RdrType] }
>    :             AtomicType                                           { [$1]    }
>    | AtomicTypes AtomicType                                           { $2 : $1 }

> AtomicType :: { RdrType }
>    : QTypeConstrID                                                    { RdrTypeConstr $1          }
>    | TypeVarID                                                        { RdrTypeVar $1             }
>    | '(' ')'                                                          { RdrTypeTuple []           }
>    | '(' TypeExpr ')'                                                 { $2                        }
>    | '(' TypeExprs2 ')'                                               { RdrTypeTuple (reverse $2) }
>    | '[' TypeExpr ']'                                                 { RdrTypeList $2            }

> TypeVarID :: { RdrTypeVarID }
>    : IDENTIFIER                                                       { RdrName $1 }

> TypeExprs2 :: { [RdrType] }
>    : TypeExpr   ',' TypeExpr                                          { [$3, $1] }
>    | TypeExprs2 ',' TypeExpr                                          { $3 : $1  }

---------- fixity declaration --------------------------------------------------------------------

> FixityDeclaration :: { RdrFixityDecl }
>    : FixityKeyword Natural InfixOpIDs                                 { RdrFixityDecl (Fixity $2 $1) (reverse $3) }

> FixityKeyword :: { Associativity }
>    : 'infixl'                                                         { LeftAssoc  }
>    | 'infixr'                                                         { RightAssoc }
>    | 'infix'                                                          { NonAssoc   }

-- SUP: We are recycling integer literals here, the small price we have to pay is a sanity check.

> Natural :: { Precedence }
>    : INT_LITERAL                                                      {% checkPrecedence $1 }

> InfixOpIDs :: { [RdrInfixOpID] }
>    :                InfixOpOrConstrID                                 { [$1]    }
>    | InfixOpIDs ',' InfixOpOrConstrID                                 { $3 : $1 }

> InfixOpOrConstrID :: { RdrInfixOpID }
>    : InfixOpID                                                        { $1 }
>    | InfixConstrID                                                    { $1 }

> InfixOpID :: { RdrInfixOpID }
>    : '-'                                                              { RdrName "-" }
>    | INFIXOPID                                                        { RdrName $1  }

---------- function declaration et al ------------------------------------------------------------

> FunctionDeclaration :: { RdrFunctionDeclaration }
>    : Signature                                                        { $1             }
>    | EvalAnnot                                                        { $1             }
>    | Equat                                                            { RdrEquation $1 }


> Signature :: { RdrFunctionDeclaration }
>    : FunctionNames '::' TypeExpr                                      { RdrSignature (reverse $1) $3 }

> FunctionNames :: { [RdrName] }
>    :                   FunctionName                                   { [ $1 ]  }
>    | FunctionNames ',' FunctionName                                   { $3 : $1 }


-- !!! varid | ( varsym ) => INFIXOP
> FunctionName :: { RdrName }
>    : '(' InfixOpID ')'                                                { $2 }
>    | FunctionID                                                       { $1 }

> FunctionID :: { RdrName }
>    : IDENTIFIER                                                       { RdrName $1 }

> EvalAnnot :: { RdrFunctionDeclaration }
>    : FunctionName 'eval' Annotation                                   { RdrEvalAnnot $1 $3 }

> Annotation :: { Annotation }
>    : 'flex'                                                           { Flex   }
>    | 'rigid'                                                          { Rigid  }
>    | 'choice'                                                         { Choice }

---------- equation ------------------------------------------------------------------------------

> Equat :: { RdrEquat }
>    : FunLHS '=' Expr  MaybeWhere                                      { RdrEquat (fst $1) (snd $1) (RdrUnguardedRHS $3) $4          }
>    | FunLHS CondExprs MaybeWhere                                      { RdrEquat (fst $1) (snd $1) (RdrGuardedRHSs (reverse $2)) $3 }
									
> MaybeWhere :: { [RdrLocalDef] }
>    : {- EMPTY -}                                                      { [] }
>    | 'where' LocalDefs                                                { $2 }

> FunLHS :: { (RdrFunctionName, [RdrPattern]) }							
>    : FunctionName MaybeSimplePatterns                                 { ($1, reverse $2) }
>    | SimplePattern InfixOpID SimplePattern                            { ($2, [$1, $3])   }
									
> MaybeSimplePatterns :: { [RdrPattern] }
>    : {- EMPTY -}                                                      { []      }
>    | MaybeSimplePatterns SimplePattern                                { $2 : $1 }

-- SUP: We allow *qualified* infix constructors here, too.

> Pattern :: { RdrPattern }
>    : FooPattern                                                       { $1                         }
>    | FooPattern QInfixConstrID Pattern                                { RdrPatternAppl $2 [$1, $3] }

-- SUP: Not very qualified yet...

> QInfixConstrID :: { RdrQDataConstrID }
>    : INFIXCONSTRUCTOR                                                 { RdrUnqualName $1 }

> FooPattern :: { RdrPattern }
>    : QDataConstrID SimplePatterns                                     { RdrPatternAppl $1 (reverse $2) }
>    | SimplePattern                                                    { $1                             }

> SimplePatterns :: { [RdrPattern] }
>    :                SimplePattern                                     { [$1]    }
>    | SimplePatterns SimplePattern                                     { $2 : $1 }


-- SUP: Distinction between VariableID and NonVarPattern because of
--      ambiguity in local definitions (variable vs. 0-ary function)

> SimplePattern :: { RdrPattern }
>    : VariableID                                                       { RdrPatternVar $1    }
>    | '_'                                                              { RdrPatternAnonymous }
>    | NonVarPattern                                                    { $1                  }

> NonVarPattern :: { RdrPattern }
>    : QDataConstrID                                                    { RdrPatternAppl $1 []           }
>    | Literal                                                          { RdrPatternLiteral $1           }
>    | '(' ')'                                                          { RdrPatternTuple []             }
>    | '(' Patterns2 ')'                                                { RdrPatternTuple (reverse $2)   }
>    | '(' Pattern ')'                                                  { $2                             }
>    | '[' MaybePatterns ']'                                            { RdrPatternList $2              }

> Patterns2 :: { [RdrPattern] }
>    : Pattern   ',' Pattern                                            { [$3, $1] }
>    | Patterns2 ',' Pattern                                            { $3 : $1  }
									
> MaybePatterns :: { [RdrPattern] }
>    : {- EMPTY -}                                                      { []      }
>    | Patterns1                                                        { reverse $1 }

> Patterns1 :: { [RdrPattern] }
>    :               Pattern                                            { [$1]    }
>    | Patterns1 ',' Pattern                                            { $3 : $1 }

> VariableID :: { RdrVariableID }						
>    : IDENTIFIER                                                       { RdrName $1 }
									
> LocalDefs :: { [RdrLocalDef] }
>    : '{' LayoutOff ValueDeclarations '}'                              { reverse $3 }
>    |     LayoutOn  ValueDeclarations Close                            { reverse $2 }
									
> ValueDeclarations :: { [RdrLocalDef] }
>    :                       ValueDeclaration                           { [$1]    }
>    | ValueDeclarations ';' ValueDeclaration                           { $3 : $1 }

-- SUP: We use FunctionNames instead of VariableIDs in the last alternative.
--      Reason: r/r conflict for "x,y,z::Int" and "x,y,z free". With finite
--      lookahead it is not decidable whether the identifiers are
--      FunctionNames or VariableIDs.

> ValueDeclaration :: { RdrLocalDef }
>    : FunctionDeclaration                                              { RdrLocalFuncDecl $1           }
>    | PatternDeclaration                                               { $1                            }
>    | FunctionNames 'free'                                             { RdrLocalFreeDecl (reverse $1) }
									
> PatternDeclaration :: { RdrLocalDef }
>    : NonVarPattern '=' Expr MaybeWhere                                { RdrLocalPatternDecl $1 $3 $4 }
									
---------- expressions ---------------------------------------------------------------------------

> CondExprs :: { [RdrGuardedRHS] }							
>    :           CondExpr                                               { [$1]    }
>    | CondExprs CondExpr                                               { $2 : $1 }
									
> CondExpr :: { RdrGuardedRHS }							
>    : '|' Expr '=' Expr                                                { RdrGuardedRHS $2 $4 }

> Expr :: { RdrExpr }
>    : '\\' SimplePatterns '->' Expr                                    { RdrExprLambda (reverse $2) $4                  }
>    | 'let' LocalDefs 'in' Expr                                        { RdrExprLet $2 $4                               }
>    | 'if' Expr 'then' Expr 'else' Expr                                { RdrExprIf $2 $4 $6                             }
>    | Expr InfixOpOrConstrID Expr                                      { RdrExprAppl (RdrExprAppl (name2expr $2) $1) $3 }
>    | '-' Expr                                                         { mkNegate $2                                    }
>    | FunctExpr                                                        { $1                                             }

> FunctExpr :: { RdrExpr }							
>    :           BasicExpr                                              { $1                }
>    | FunctExpr BasicExpr                                              { RdrExprAppl $1 $2 }

-- SUP: In the last alternative we use INFIXOPID/QInfixConstrID instead of
--      QInfixOpOrConstrID because (-x) is not a right section, but an
--      expression in parens.

> BasicExpr :: { RdrExpr }
>    : QVariableID                                                      { RdrExprVar $1                      }
>    | Literal                                                          { RdrExprLiteral $1                  }
>    | '(' QInfixOpOrConstrID ')'                                       { RdrExprVar $2                      }
>    | QDataConstrID                                                    { RdrExprConstr $1                   }
>    | '(' ')'                                                          { RdrExprTuple []                    }
>    | '(' Expr ')'                                                     { $2                                 }
>    | '(' Exprs2 ')'                                                   { RdrExprTuple (reverse $2)          }
>    | '[' MaybeExprs ']'                                               { RdrExprList $2                     }
>    | '(' Expr QInfixOpOrConstrID ')'                                  { leftSection $2 $3                  }
>    | '(' INFIXOPID Expr ')'                                           { rightSection (RdrUnqualName $2) $3 }
>    | '(' QInfixConstrID Expr ')'                                      { rightSection $2 $3                 }

> QInfixOpOrConstrID :: { RdrQName }
>    : InfixOpID                                                        { case $1 of RdrName n -> RdrUnqualName n }
>    | QInfixConstrID                                                   { $1                                      }

> Exprs2 :: { [RdrExpr] }
>    : Expr   ',' Expr                                                  { [$3, $1] }
>    | Exprs2 ',' Expr                                                  { $3 : $1  }
									
> MaybeExprs :: { [RdrExpr] }
>    : {- EMPTY -}                                                      { []         }
>    | Exprs                                                            { reverse $1 }
									
> Exprs :: { [RdrExpr] }							
>    :           Expr                                                   { [$1]    }
>    | Exprs ',' Expr                                                   { $3 : $1 }
									
> QTypeConstrID :: { RdrQName }						
>    : DOTTED_CONSTRUCTOR                                               { RdrQualName (RdrModuleID (init $1)) (last $1) }
>    | CONSTRUCTOR                                                      { RdrUnqualName $1 }
									
> QDataConstrID :: { RdrQName }						
>    : DOTTED_CONSTRUCTOR                                               { RdrQualName (RdrModuleID (init $1)) (last $1) }
>    | CONSTRUCTOR                                                      { RdrUnqualName $1 }
									
> QVariableID :: { RdrQName }							
>    : DOTTED_IDENTIFIER                                                { RdrQualName (RdrModuleID (init $1)) (last $1) }
>    | IDENTIFIER                                                       { RdrUnqualName $1 }
									
> Literal :: { Literal }
>    : INT_LITERAL                                                      { LiteralInt (read $1) }

---------- layout --------------------------------------------------------------------------------

> Close :: { () }
>    : VIRTUALBRACE                                                     { () } -- context popped in lexer.
>    | error                                                            {% popContext }

> LayoutOff :: { () }
>    : {- EMPTY -}                                                      {% pushContext NoLayout }

> LayoutOn  :: { () }
>    : {- EMPTY -}                                                      {% getSrcLoc `thenP` \(SrcLoc r c) ->
>                                                                          pushContext (Layout c) }

---------- Here comes the HORRIBLE multiple entries hack... --------------------------------------

> {
> data ParseType = ItsAModule RdrModule
>                | ItsAnExpression RdrExpr

> parseModule :: String -> MaybeError RdrModule
> parseModule input = case parseMOrE ('|':'\n':input) of
>                        Ok _ (ItsAModule m)      -> return m
>                        Ok _ (ItsAnExpression _) -> error "impossible (parseModule)"
>                        Failed err               -> fail err
> 
> parseExpression :: String -> MaybeError RdrExpr
> parseExpression input = case parseMOrE (';':'\n':input) of
>                            Ok _ (ItsAModule _)      -> error "impossible (parseExpression)"
>                            Ok _ (ItsAnExpression e) -> return e
>                            Failed err               -> fail err

-- We start at line 0 because of the multiple entries hack

> parseMOrE input = parseCurry input (SrcLoc 0 1) 0 []

---------- error "handling" ----------------------------------------------------------------------

> happyError = parseError "Parse error"

---------- utility functions ---------------------------------------------------------------------

> checkPrecedence :: String -> P Precedence
> checkPrecedence cs@[c] | isDigit c = returnP (Precedence (read cs))
> checkPrecedence cs                 = parseError ("Illegal precedence " ++ cs)

> name2expr :: RdrName -> RdrExpr
> name2expr (RdrName n) = RdrExprVar (RdrUnqualName n)

> splitBasicType :: RdrType -> P RdrConstrDecl
> splitBasicType t = split t []
>    where split (RdrTypeAppl t u)                 ts = split t (u:ts)
>          split (RdrTypeConstr (RdrUnqualName c)) ts = returnP (RdrConstrDecl (RdrName c) ts)
>          split _                                 _  = parseError "Illegal data declaration"

> mkNegate :: RdrExpr -> RdrExpr
> mkNegate = RdrExprAppl (RdrExprVar (RdrUnqualName "negate"))

> leftSection :: RdrExpr -> RdrQName -> RdrExpr
> leftSection expr op = RdrExprAppl (RdrExprVar op) expr

> rightSection :: RdrQName -> RdrExpr -> RdrExpr
> rightSection op expr = RdrExprLambda [RdrPatternVar v] (RdrExprAppl (RdrExprAppl (RdrExprVar op) (name2expr v)) expr)
>    where v = RdrName "_rsec"

> }
