module PrettyCurryCoreSyntax where

import List(intersperse)
import Pretty
import CommonSyntax
import CurryCoreSyntax



instance Pretty CCExpr where
   pretty s (CCExprLet ls e)         = parens (text "let" $$ nest 4 (vcatP s ls) $$
                                               text "in " <+> pretty s e)
   pretty s (CCExprFreeDecl v e)     = parens (text "letfree" <+> pretty s v <+> pretty s e)
   pretty s (CCExprLambda v e)       = parens (text "\\ " <> pretty s v <+> text "->" <+> pretty s e)
   pretty s (CCExprCase a v alts)    = parens (text ("case") <+> pretty s a <+>
					       pretty s v <+> text "of" $$ nest 3 (vcatP s alts))
   pretty s e@(CCExprAppl _ _)       = parens (foo e)
      where foo (CCExprAppl f v) = foo f <+> pretty s v
            foo x                = pretty s x
   pretty s (CCExprVar v)            = pretty s v

   pretty s (CCExprConstr c vars)    = pretty s c <+> sepP s vars
   pretty s (CCExprStrict expr)	     = parens (text "strict" <+> pretty s expr)
   pretty s e@(CCExprAnd _ _)        = parens (foo e)
      where foo (CCExprAnd e1 e2) = foo e1 <+> text "&" <+> foo e2
            foo x                 = pretty s x
   pretty s e@(CCExprOr _ _)         = parens (foo e)
      where foo (CCExprOr e1 e2)  = foo e1 <+> text "or" <+> foo e2
            foo x                 = pretty s x
   pretty s (CCExprUnify e1 e2)	     = parens (pretty s e1 <+> text "=:=" <+> pretty s e2)
   pretty s (CCExprGuard e1 e2)	     = parens (pretty s e1 <+> text "|" <+> pretty s e2)
   pretty _ CCExprFail		     = text "_fail"
   pretty _ (CCExprError _)          = text "_error" 

instance Pretty CCFunDef where
   pretty s (CCFunDef name body) =
      pretty s name <+> equals <+> pretty s body

instance Pretty CCAlternative where
   pretty s (CCAlternative c vs e) = pretty s c <+> sepP s vs <+> text "->" <+> pretty s e

instance Pretty CCVariable where
   pretty _ (CCVariable id) = text "v_" <> int id

instance Pretty CCDataConstr where
   pretty s (CCDataConstr name id _) = pretty s name <> text "[" <> int id <> text "]"

instance Pretty CCName where
   pretty _ (CCName name) = text name
