module PrettyTypeCheckerSyntax where

import List(intersperse)
import Pretty
import CommonSyntax
import TypeCheckerSyntax

---------- module structure ----------------------------------------------------------------------

instance Pretty TcModule where
   pretty s (TcModule ms ds ls) = text "pragma" <+> pretty s ms $$ vcatP s ds $$ vcatP s ls

---------- declaration of a new algebraic type ---------------------------------------------------

instance Pretty TcDataDeclaration where
   pretty s (TcDataDeclaration c vs cs) =
      text "data" <+> pretty s c <+> sepP s vs <+> equals <+>
      (hcat . intersperse (text " | ") . map (pretty s)) cs $$
      text ""

instance Pretty TcConstrDecl where
   pretty s (TcConstrDecl c ts) = pretty s c <+> sepP s ts

---------- type ----------------------------------------------------------------------------------

instance Pretty TcType where
   pretty s (TcTypeConstr c []) = pretty s c -- SUP: special case, only for nice layout
   pretty s (TcTypeConstr c ts) = pretty s c <> braces (sepByCommas s ts)
   pretty s (TcTypeVar v)       = pretty s v

---------- local definition, equation ------------------------------------------------------------

instance Pretty TcLocalDef where
   pretty s (TcLocalFuncDecl mbTy mbAnn n eqs) =
      text "===== function" <+> pretty s n <+>
      text "::" <+> maybe (text "???") (pretty s) mbTy <> text "," <+>
      text "eval" <+> maybe (text "???") (pretty s) mbAnn <+> text "=====" $$
      vcatP s eqs $$
      text ""
   pretty s (TcLocalPatternDecl p e) =
      text "===== pattern =====" $$
      pretty s p <+> equals <+> pretty s e $$
      text ""

instance Pretty TcEquat where
   pretty s (TcEquat ps expr) = sepP s ps <+> equals <+> pretty s expr

instance Pretty TcPattern where
   pretty s (TcPatternVar v)     = pretty s v
   pretty _ TcPatternAnonymous   = text "_"
   pretty s (TcPatternLiteral l) = pretty s l
   pretty s (TcPatternAppl c []) = pretty s c   -- SUP: special case, only for nice layout
   pretty s (TcPatternAppl c ps) = parens (pretty s c <+> sepP s ps)

---------- expressions ---------------------------------------------------------------------------

instance Pretty TcExpr where
   pretty s (TcExprLambda   ps e) = parens (text "\\" <> sepP s ps <+> text "->" <+> pretty s e)
   pretty s (TcExprFreeDecl vs e) = parens (text "let" <+> sepByCommas s vs <+> text "free" $$
                                            text "in" <+> pretty s e)
   pretty s (TcExprLet      l  e) = parens (text "let" <+> pretty s l $$
                                            text "in" <+> pretty s e)
   pretty s (TcExprLetrec   ls e) = parens (text "letrec" $$ nest 4 (vcatP s ls) $$
                                            text "in" <+> pretty s e)
   pretty s e@(TcExprAppl _ _)    = parens (foo e)
       where foo (TcExprAppl f a) = foo f <+> pretty s a
             foo x                = pretty s x
   pretty s (TcExprVar v)
                 | isOperatorTc v = parens (pretty s v)
                 | otherwise      = pretty s v
   pretty s (TcExprConstr c)      = pretty s c
   pretty s (TcExprLiteral l)     = pretty s l

---------- names ---------------------------------------------------------------------------------

instance Pretty TcQName where
   pretty s (TcQualName m n) = pretty s m <> text ('.':n)
   pretty _ (TcUnqualName n) = text n

instance Pretty TcName where
   pretty _ (TcName n) = text n

instance Pretty TcModuleID where
   pretty _ (TcModuleID ss) = text . concat . intersperse "." $ ss
