module TypeCheckerSyntax where

import Char(isAlpha)
import CommonSyntax(MatchStrategy, Literal, Annotation)

data TcModule =
     TcModule MatchStrategy
	      [TcDataDeclaration]
              [TcLocalDef]                                         -- only TcLocalFuncDecls

---------- declaration of a new algebraic type ---------------------------------------------------

data TcDataDeclaration =
     TcDataDeclaration TcTypeConstrID [TcTypeVarID] [TcConstrDecl]
     deriving (Show) -- for debugging purposes

data TcConstrDecl =
     TcConstrDecl TcDataConstrID [TcType]
     deriving (Show) -- for debugging purposes

---------- type ----------------------------------------------------------------------------------

data TcType =
     TcTypeConstr TcQTypeConstrID [TcType]
   | TcTypeVar TcTypeVarID
     deriving (Show,Ord,Eq) -- for debugging purposes

---------- local definition, equation ------------------------------------------------------------

data TcLocalDef =
     TcLocalFuncDecl    (Maybe TcType) (Maybe Annotation) TcFunctionName [TcEquat]
   | TcLocalPatternDecl TcPattern TcExpr
     deriving (Show,Ord,Eq) -- for debugging purposes

data TcEquat =
     TcEquat [TcPattern] TcExpr
     deriving (Show,Ord,Eq) -- for debugging purposes

data TcPattern =
     TcPatternVar       TcVariableID
   | TcPatternAnonymous
   | TcPatternLiteral   Literal
   | TcPatternAppl      TcQDataConstrID [TcPattern]
     deriving (Show,Ord,Eq) -- for debugging purposes

---------- expressions ---------------------------------------------------------------------------

data TcExpr =
     TcExprLambda   [TcPattern] TcExpr
   | TcExprFreeDecl [TcVariableID] TcExpr
   | TcExprLet      TcLocalDef TcExpr
   | TcExprLetrec   [TcLocalDef] TcExpr
   | TcExprAppl     TcExpr TcExpr
   | TcExprVar      TcQVariableID   -- NOTE: Could be a function, too
   | TcExprConstr   TcQDataConstrID
   | TcExprLiteral  Literal
     deriving (Show,Ord,Eq) -- for debugging purposes

---------- names ---------------------------------------------------------------------------------

data TcQName =
     TcQualName   TcModuleID String
   | TcUnqualName String
   deriving (Eq, Ord, Show) -- for collectDataDecls

isOperatorTc :: TcQName -> Bool
isOperatorTc n = case n of
                    TcQualName _ v -> isOperatorPrefix v
                    TcUnqualName v -> isOperatorPrefix v
   where isOperatorPrefix ('_':c:_) = not (isAlpha c) -- special case for internal identifiers
         isOperatorPrefix (c:_)     = not (isAlpha c)
         isOperatorPrefix []        = error "impossible (isOperatorTc)"

newtype TcName = TcName String   
	deriving (Eq, Ord, Show) -- for checkLinearity

newtype TcModuleID = TcModuleID [String] 
	deriving (Eq, Ord, Show) -- for collectDataDecls

type TcQVariableID   = TcQName
type TcVariableID    = TcName

type TcQDataConstrID = TcQName
type TcDataConstrID  = TcName

type TcQTypeConstrID = TcQName
type TcTypeConstrID  = TcName

type TcFunctionName  = TcName

type TcTypeVarID     = TcName
