-- Database programming in Curry: family relationships
-- (functional logic style with explicit functional dependencies)

{- Structure of the family:

                    Christine --- Antony  Maria --- Bill
                      /    \              |
                     /      \             |
       Monica --- John       Alice --- Frank
        /  \                   |
       /    \                  |
    Susan  Peter             Andrew
-}

-- default evaluation strategy: narrowing, i.e., allow inversion of functions:
pragma flex


data Person = Christine | Maria | Monica | Alice | Susan |
              Antony | Bill | John | Frank | Peter | Andrew

success = 1 =:= 1

husband Christine  = Antony
husband Maria      = Bill
husband Monica     = John
husband Alice      = Frank


mother John    = Christine
mother Alice   = Christine
mother Frank   = Maria
mother Susan   = Monica
mother Peter   = Monica
mother Andrew  = Alice

father c  = husband (mother c)

grandfather g c | g =:= father (father c) = success
                | g =:= father (mother c) = success

-- A is ancestor of P:
ancestor  a p | a =:= father p = success
              | a =:= mother p = success
              | let p1 free in (a =:= father p1) & ancestor p1 p = success
              | let p1 free in (a =:= mother p1) & ancestor p1 p = success


-- Example goals:
-- main = father Peter
--main  = father child =:= John
--main   = grandfather g c
--main     = ancestor a Andrew
main   = ancestor a p

