{ -- -*- haskell -*-
module Parser(parse)
where

import Scanner(Token(..))
} 


%name parse
%tokentype { Token }
%token  "("      { TokenLParen         }
        ")"	 { TokenRParen	       }
        "{"	 { TokenLCBracket      }
        "}"	 { TokenRCBracket      }
        "["	 { TokenLSBracket      }
        "]"	 { TokenRSBracket      }
	";"	 { TokenSemicolon      }
	":"	 { TokenColon	       }
	","	 { TokenComa	       }
        "="	 { TokenEqu	       }    
	"->"	 { TokenTo	       }     
	"lambda" { TokenLambda	       }    
	"let"    { TokenLet	       }
	"in"     { TokenIn	       }
	"letfree"{ TokenLetFree	       }
	"case"	 { TokenCase	       }
	"flex"	 { TokenFlex	       }
	"rigid"	 { TokenRigid	       }
	"of"	 { TokenOf	       }
	"try"	 { TokenTry	       }
	"or"	 { TokenOr	       }
	"|"	 { TokenGuard	       }
	"&"	 { TokenAnd	       }
	"=:="	 { TokenUnify	       }
	C	 { TokenConstructor $$ }
	V	 { TokenVar $$	       }	  
%%


Expr :: { Expr }
    : "let" "{" FunDefs "}" "in" Expr	    { LetExpr      $3 $6          }
    | "letfree" V Expr			    { LetFreeExpr  (Var $2) $3    }
    | "lambda" V Expr			    { LambdaExpr   (Var $2) $3    }
    | "case" V "of" "{" Alts "}"	    { CaseExpr	   $2 (Var $3) $6 }
    | FunctExpr				    { $1 }


FunDefs :: { Functions }
    : FunDef				    { [$1] }
    | FunDef ";" FunDefs		    { $1 : $3 }

FunDef :: { Function }
    : V "=" Expr			    { Function (Var $1) $3 }


Alts :: { Alternatives }
    : Alt				    { [$1] }
    | Alt ";" Alts			    { $1 : $3 }

Alt :: { Alternative }
    : C Vars "->" Expr			    { Alternative (Constr $1) $2 $4 }

Vars :: { Vars }
    : 					    { [] }
    | V Vars				    { (Var $1) : $2 }

FunctExpr :: { Expr }
    : BasicExpr				    { $1 }
    | FunctExpr V			    { AppExpr $1 (Var $2) }

BasicExpr :: { Expr }
    : V					    { VarExpr (Var $1) }
    | C					    { ConstrExpr (Constr $1) }
    | "(" Expr ")"			    { $2 }



{}
