module CreateTex
where

import ParserDefinition(Definitions)
import FiniteMap
import ParserHappy


escapeChar '&' = True
escapeChar '_' = True
escapeChar '%' = True
escapeChar '{' = True
escapeChar '}' = True
escapeChar  _  = False

escape c = ['\\', c]

makeTexable s = replaceBy escapeChar escape s 

replaceBy toReplace replaceWith list = concatMap (\e -> if (toReplace e) then (replaceWith e) else [e]) list 


verbchars = "+!#$&*|"
terminalrep s = "\\verb" ++ [verbchar] ++ s ++ [verbchar]
	    where verbchar = head $ filter (\c -> not $ elem c s) verbchars


terminal2tex :: Definitions -> String -> String
-----------------------------------------------
terminal2tex defs s = lookupWithDefaultFM defs (terminalrep s) s


name2String :: (String -> String) -> (Name -> Bool) -> Name -> String
---------------------------------------------------------------------
name2String name2Tex isTerm (Name name)
	    | isTerm (Name name) = name2Tex name
	    | otherwise		 = makeTexable name 


createTex :: Definitions -> HappyFile -> String -> Int -> IO(String)
-------------------------------------------------------------
createTex defs (HappyFile terminals rules) macroName maxlength
  | null macroName = return $ "\\begin{longtable}{lcl}\n" ++
			      createTex' rules ++
			      "\\end{longtable}\n"
  | otherwise	   = return $ "\\def\\" ++ macroName ++
			      "{\\begin{longtable}{lcl}\n" ++
			      createTex' rules ++
			      "\\end{longtable}}\n"
  where  toTex  = terminal2tex defs
	 isTerm = isTerminal terminals 
         createTex' [] = []
         createTex' ((Rule (Name name) alts):rules) =
	    toLength (makeTexable name) ++ "& ::= & " ++ createAlts 0 alts ++
	    createTex' rules
         createAlts _ [(Alternative names)] =
            createAlt names ++ ".\\\\\n"
         createAlts n ((Alternative names):alts) =
            createAlt names ++ " $|$ " ++ nl ++ "\n" ++
            toLength "" ++  sp   ++ createAlts n'' alts
	    where n'  = n + (sum $ map length [n | (Name n) <- names])
		  (n'', nl, sp) = if maybeNL n' alts then (n', "", "        ") 
						     else (0, "\\\\", "&     & ")
         createAlt [] = "\\ensuremath{\\epsilon}"
	 createAlt names = createAlt' names
	 createAlt' [] = []
         createAlt' (name:names) = 
            name2String toTex isTerm name ++ " " ++ createAlt' names
	 toLength s = s ++ (take (25-(length s)) (repeat ' '))
	 maybeNL n ((Alternative names):_) = 
	      n + (sum $ map length [n' | (Name n') <- names])  < maxlength
	      
