module Main
where

import ReadHappy
import DefinitionFM
import FiniteMap
import IO
import CreateTex
import GetOpt
import Char

#if __HUGS__
getArgs :: IO [String]
getArgs = return []
#else
import System
#endif

data Option = WriteFile String | DefinitionFile String | MacroName String | LineLength String | Help
     deriving Eq

options :: [OptDescr Option]
options =
   [Option ['w'] ["write"]       (ReqArg WriteFile      "FILE")   "Write output into FILE",
    Option ['d'] ["definition"]  (ReqArg DefinitionFile "FILE")   "Use definitionfile FILE",
    Option ['m'] ["macro"]	 (ReqArg MacroName	"NAME")   "Define the grammar by the TeX-Macro NAME",
    Option ['l'] ["length"]	 (ReqArg LineLength	"L")	  "L is the maximal length of a rules right side",
    Option ['h'] ["help"]        (NoArg  Help)		"describe options"]


usage = putStr (usageInfo 
	       ("\nUsage: happy2latex [OPTION...] happyfile" ++ 
	        "\n       (literate happy is not yet supported!)\n" ++
                "where OPTION is one of:") options)

mainHugs args = handleCmdline (getOpt Permute options args)
   where handleCmdline (opts,fileNames,[]  ) | Help `elem` opts = usage
                                             | otherwise        = go opts fileNames
         handleCmdline (_   ,_        ,errs)			= do putStr (concat errs)
		

main = do args <- getArgs
	  mainHugs args			

go options filenames = 
   do hf  <- createHappyFile (head filenames)
      dFM <- createDFM options
      tex <- createTex dFM hf (macroName options) (lineLength options)
      writeOut options tex



createDFM options
   | isOptDefinition = do createDefinitionFM $ removeSpaces (head definitions)
   | otherwise	     = return emptyFM
   where definitions     = [fn | (DefinitionFile fn) <- options]
	 isOptDefinition = not $ null definitions

writeOut options tex
   | isOptWriteFile  = let outFile   = removeSpaces $ head outfiles
		       in do out <- openFile outFile WriteMode
			     hPutStr out tex
			     hClose out
   | otherwise       = do putStr tex
   where outfiles       = [fn | (WriteFile fn) <- options] 
	 isOptWriteFile = not $ null outfiles


macroName options
   | null macroNames = []
   | otherwise	     = removeSpaces $ head macroNames
   where macroNames = [mn | (MacroName mn) <- options]

lineLength options
   | null ll = 0
   | otherwise = stringToInt $ removeSpaces $ head ll
   where ll = [i | (LineLength i) <- options]
	 
stringToInt s = sum $ map (\(x, y) -> x*y) (zip (map digitToInt (reverse s)) dec)
	    where dec = 1:(map (\x -> x*10) dec)

removeSpaces s = filter (\c -> not $ c == ' ') s