-- -*- haskell -*- 
{  
module ParserHappy
where

import Scanner(Token(..))
import Set
} 


%name parseHappy
%tokentype { Token }
%token  "%%"          { TokenStart          }
	"{...}"	      { TokenHaskell        }
        "happyname"   { TokenHappyName      }
	"token"       { TokenHappyToken	    }
	UNKNOWN	      { TokenHappyUnknown   }
	":"	      { TokenBeBeComes	    }
	"|"	      { TokenAlternative    }
	NAME	      { TokenName	 $$ }  
%%


definitionfile :: { HappyFile }
   : maybeCode dirs "%%" rules "{...}"         { HappyFile (mkSet $2) $4 }

maybeCode :: {}
   : "{...}" {}
   |         {}

maybeCodes :: {}
   :                    {}
   | "{...}" maybeCodes {}

dirs :: { Names }
   :		{ [] }
   | dir dirs   { $1 ++ $2 }

dir :: { Names }
   : "happyname" name		 { [] }
   | UNKNOWN maybeCodes	         { [] }
   | "token" tokens		 { $2 }

tokens :: { Names }
    : token		    { [$1] }
    | token tokens	    { $1 : $2 }

token :: { Name }
    : name "{...}" { $1 }

rules :: { Rules }
   :                       { [] }
   | rule rules		   { $1 : $2 }

rule :: { Rule }
   : name ":" ":" "{...}" ":" alternatives { Rule $1 $6 }

alternatives :: { Alternatives }
   : alternative	          { [$1] }
   | alternative "|" alternatives { $1 : $3 }

alternative :: { Alternative }
   : names "{...}"                { Alternative $1}

names :: { Names }
   :				  { [] }
   | name names			  { $1 : $2 }

name :: { Name }
   : NAME			  { Name $1 }

{ 
data HappyFile = HappyFile Terminals Rules

type Terminals = Set Name

type Names = [Name]

data Name = Name String
     deriving (Eq,Ord)

type Rules = [Rule]

data Rule = Rule Name Alternatives

type Alternatives = [Alternative]

data Alternative = Alternative Names

isTerminal :: Terminals -> Name -> Bool
---------------------------------------
isTerminal ts name = elementOf name ts

parseHappy' ts = return $ parseHappy ts 

happyError t = error $ "ParseHappy: Fehler nahe " ++ (show (take 5 t)) ++ "\n"
}
