-- -*- haskell -*-
%{
module Scanner(Token(..), scanner')
where

import Alex(load_scan, scan, Posn)
%}

{ ^s = [^t^v^f^r^ ] }  -- spaces
{ ^b = [a-zA-Z]     } 
{ ^a = [a-zA-Z0-9^_] }
{ ^e = [^9^n^32-^122^124^126] } -- everything but { and }
{ ^h = [^9^n^32-^44^46-^122^124^126] } -- everything but '-' and { and }
{ ^t = [^9^32-^126] }       -- everything but newline
{ ^q = [^9^32-^126]#[^34] } -- everything but newline and "
{ ^r = [^9^32-^126]#[^39] } -- everything but newline and '


"table"/"actions" :-

<>         ::= ^n           -- Newline 
<>         ::= ^s+          -- Whitespace
<>         ::= ^-^-.*       -- Comments
<>	   ::= ^{^-^e*^-^}
--                                     Position String
--                                            v v
<haskell> ::= ^{^h^e*^h^}|^{^}|^{^h^}       %{ haskell    _ _ = TokenHaskell              %} 
<colon>	  ::= ^:		%{ colon      _ _ = TokenBeBeComes	      %} 
<alt>	  ::= ^|		%{ alt        _ _ = TokenAlternative	      %}
<name>	  ::= (^b^a*)|(^34^q*^34)|(^39^r*^39)	%{ name       _ s = TokenName s		      %}
<happy>	  ::= ^%^a*		%{ happy      _ s = createHappyToken s	      %}
<start>	  ::= ^%^%		%{ start      _ _ = TokenStart		      %}
<texterm> ::= ^-^>^t*^n		%{ texterm    _ s = TokenTexRepresentation s  %}

%{
data Token = 
	     TokenStart
	   | TokenHaskell
	   | TokenHappyName
	   | TokenHappyToken
	   | TokenHappyUnknown

 	   | TokenBeBeComes
	   | TokenAlternative
	   | TokenName               String
	   | TokenTexRepresentation  String

	   | TokenError       Posn   String
     deriving (Show)


createHappyToken :: String -> Token
createHappyToken "%name"      = TokenHappyName
createHappyToken "%token"     = TokenHappyToken
createHappyToken s	      = TokenHappyUnknown

scanner' c = return $ scanner c

scanner :: String -> [Token]
scanner = scan (load_scan (actions, stop_action) table)
   where stop_action _ ""   = []
         stop_action p rest = [TokenError p rest]
%}
