% -*- prolog -*-

% Semantische Operatoren
:- op(900, xfy, [--->>, --->>*]).

:- ['Utilities'].
:- ['VarSupp'].

:- ['Syntax'].
:- ['Substitute'].

:- ['Heap'].
:- ['OpSem'].

:- ['Examples'].




% --->>

([Config|Leaves], Closed) --->> Out :-
   ( Config ---> Configs ->
       myAppend(Configs, Leaves, NewLeaves),
       Out = (NewLeaves, Closed)
       ;   
       Out = (Leaves, [Config|Closed]) ).

%([Config|In],Out) --->> (In_, Out) :-
%   Config ---> Configs,
%   myAppend(Configs, In, In_).
%([Config|In],Out) --->> (In, [Config|Out]) :-
%   \+ (Config ---> _).

% --->>*
Configs --->>* Configs_ :-
   Configs --->>  Cs, !,
   Cs      --->>* Configs_.
Configs --->>* Configs.

% Auswertungsrelation eval

eval(Config, Configs) :-
   ([Config],[]) --->>* ([], Configs).

% Initiale Konfiguration m_i

resVar(v(result)).

initHeap(E, Free, Heap) :-
   emptyHeap(Heap_),
   freeVars(E, Free),
   initializeVars(Heap_, Free, Heap__),
   resVar(P),
   addBinding(Heap__, [P, E], Heap).

m_i(E, Free, config(strict(P), Heap)) :-
   initHeap(E, Free, Heap),
   resVar(P).
   

% Semantische Relation sem

sem(E, R) :-
   freeVars(E, Free),
   m_i(E, Free, M_i),
   eval(M_i, M),
   semSub(M, Free, R).

semSub([], _, []).
semSub([config(constr(success,[]), Heap) |M], Free, [(SubstFree, Result) |R]) :-
   !,
   inducedSubst(Heap, HeapSigma),
   reduceSubst(HeapSigma, Free, SubstFree_),
   substIter(SubstFree_, SubstFree),
   resVar(P),
   reduceSubst(HeapSigma, [P], SubstRes_),
   substIter(SubstRes_, SubstRes),
   substitute(P, SubstRes, Result),
   semSub(M, Free, R).
semSub([_|M], Free, R) :-
   semSub(M, Free, R).

% Zum Ausfhren der Beispiele

runExample(N) :-
   initVarSupp, 
   prog(N, E), 
   sem(E,R),
   showAnswerExpressions(R).

% Ausgabe der Antwortausdrcke

showAnswerExpressions([]) :-
   !, nl, write('No Solutions.'), nl.
showAnswerExpressions(R) :-
   showAnswerExpressionsSub(R).

showAnswerExpressionsSub([]).
showAnswerExpressionsSub([AE|R]) :-
   showAnswerExpression(AE),
   showAnswerExpressionsSub(R).

showAnswerExpression((Subst, Term)) :-
   nl, write('------ answer expression ------'), nl,
   showCTerm(Term),
   nl,
   nl,
   showSubst(Subst).

showSubst([]).
showSubst([[X,X]|Subst]) :-
   !, showSubst(Subst).
showSubst([[X,T]|Subst]) :-
   write(X),
   write('\t=\t'),
   showCTerm(T),
   nl,
   showSubst(Subst).

   