% -*- prolog -*- 

% semantical operator --->
:- op(900, xfy, [--->]).


% let
config(let(Defs,E), Heap) ---> [config(E_, Heap_)] :-
   zip(Vars, Es, Defs),
   length(Vars, N),
   newVars(N, Ps),
   zip(Vars, Ps, Subst),
   substituteList(Es, Subst, Es_),
   substitute(E, Subst, E_),
   zip(Ps, Es_, Bindings),
   addBindings(Heap, Bindings, Heap_).


% letfree
config(letfree(X, E), Heap) ---> [config(E_, Heap_)] :-
   newVar(P),
   substitute(E, [[X,P]], E_),
   initializeVar(Heap, P, Heap_).

% app
config(app(E,P),Heap) ---> Configs :-
   config(E,Heap) ---> Configs_,
   composeConfigs(Configs_, (app(E_,P),E_), Configs).
config(app(lambda(X,E),P),Heap) ---> [config(E_,Heap)] :-
   substitute(E, [[X,P]], E_).

% var
config(v(P),Heap) ---> Configs :-
   lookupBinding(Heap, v(P), E),
   removeBinding(Heap, v(P), Heap_),
   config(E,Heap_) ---> Configs_,
   allVars(Configs_, v(P), Configs).
config(v(P),Heap) ---> [config(E,Heap)] :-
   lookupBinding(Heap, v(P), E),
   whnf(E,Heap).

% case
config(case(A,E,Alts), Heap) ---> Configs :-
   config(E,Heap) ---> Configs_,
   composeConfigs(Configs_, (case(A,E_,Alts), E_),Configs).
config(case(_,constr(C,Ps),Alts), Heap) ---> [config(E_,Heap)] :-
   myMember((constr(C,Xs),E), Alts),
   zip(Xs,Ps,Subst),
   substitute(E,Subst,E_).
config(case(_,constr(C,_),Alts), _) ---> [] :-
   \+ myMember((constr(C,_),_), Alts).
config(case(flex,v(P),Alts), Heap) ---> Configs :-
   lookupBinding(Heap, v(P), nil),
   narrow(v(P), Alts, Heap, Configs).

% guard
config(G=>E,Heap) ---> Configs :-
   config(G,Heap) ---> Configs_,
   composeConfigs(Configs_, (G_ => E, G_), Configs).
config(constr(success,[])=>E, Heap) ---> [config(E,Heap)].


% &
config(E1 & E2, Heap) ---> Configs :-
   config(E1,Heap) ---> Configs_,
   composeConfigs(Configs_, (E2 & E1_, E1_), Configs).
config(E1 & E2, Heap) ---> Configs :-
   config(E2,Heap) ---> Configs_,
   composeConfigs(Configs_, (E2_ & E1, E2_), Configs).
%config(E1 & E2, Heap) ---> Configs :-
%   selectWhich([E1,E2], Which, Alternative),
%   (config(Which,Heap) ---> Configs_   ->
%       composeConfigs(Configs_, (Which_ & Alternative, Which_),Configs)
%       ;
%       config(Alternative,Heap) ---> Configs__,
%       composeConfigs(Configs__, (Which & Alternative_, Alternative_),Configs)).
config(constr(success,[]) & constr(success,[]), Heap) ---> [config(constr(success,[]),Heap)].


%selectWhich(Es,Which,Alternative) :- 
%   random(1,3,WhichI),
%   nth(WhichI, Es, Which),
%   myDelete(Es, Which,  [Alternative]).


% =:=
config(E1 =:= E2, Heap) ---> Configs :-
   config(E1,Heap) ---> Configs_,
   composeConfigs(Configs_, (E1_ =:= E2, E1_),Configs).
config(E1 =:= E2, Heap) ---> Configs :-
   config(E2,Heap) ---> Configs_,
   composeConfigs(Configs_, (E1 =:= E2_, E2_),Configs).
%config(E1 =:= E2, Heap) ---> Configs :-
%   selectWhich([E1,E2], Which, Alternative),
%   (config(Which,Heap) ---> Configs_   ->
%       composeConfigs(Configs_, (Which_ =:= Alternative, Which_),Configs)
%       ;
%       config(Alternative,Heap) ---> Configs__,
%       composeConfigs(Configs__, (Which =:= Alternative_, Alternative_),Configs)).
config(v(P1) =:= v(P2), Heap) ---> [config(constr(success,[]),Heap_)] :-
   P1 \== P2,
   lookupBinding(Heap,v(P1),nil),
   lookupBinding(Heap,v(P2),nil),
   addBinding(Heap, [v(P1),v(P2)], Heap_).
config(v(P) =:= v(P), Heap) ---> [config(constr(success,[]),Heap)] :-
   lookupBinding(Heap,v(P),nil).
config(v(P) =:= constr(C,[]), Heap) ---> [config(constr(success,[]),Heap_)] :-
   lookupBinding(Heap,v(P),nil),
   addBinding(Heap,[v(P),constr(C,[])],Heap_).
config(constr(C,[]) =:= v(P), Heap) ---> [config(constr(success,[]),Heap_)] :-
   lookupBinding(Heap,v(P),nil),
   addBinding(Heap,[v(P),constr(C,[])],Heap_).
config(v(P) =:= constr(C,Qs), Heap) ---> [config(E,Heap_)] :-
   lookupBinding(Heap,v(P),nil),
   \+ myMember(v(P),Qs),
   length(Qs,N),
   newVars(N,Rs),
   initializeVars(Heap,Rs,Heap__),
   addBinding(Heap__,[v(P),constr(C,Rs)],Heap_),
   propagateEquality(Rs,Qs,E).
config(constr(C,Qs) =:= v(P), Heap) ---> [config(E,Heap_)] :-
   lookupBinding(Heap,v(P),nil),
   \+ myMember(v(P),Qs),
   length(Qs,N),
   newVars(N,Rs),
   initializeVars(Heap,Rs,Heap__),
   addBinding(Heap__,[v(P),constr(C,Rs)],Heap_),
   propagateEquality(Rs,Qs,E).
config(v(P) =:= constr(_,Qs), Heap) ---> [] :-
   lookupBinding(Heap,v(P),nil),
   myMember(v(P),Qs).
config(constr(_,Qs) =:= v(P), Heap) ---> [] :-
   lookupBinding(Heap,v(P),nil),
   myMember(v(P),Qs).
config(constr(C,[]) =:= constr(C,[]), Heap) ---> [config(constr(success,[]),Heap)] :- !.
config(constr(C,Ps) =:= constr(C,Qs), Heap) ---> [config(E,Heap)] :-
   propagateEquality(Ps,Qs,E).
config(constr(C,_) =:= constr(D,_), _) ---> [] :-
   C \== D.

% or
config(E1 or E2, Heap) ---> [config(E1, Heap), config(E2, Heap)].


% strict
config(strict(E), Heap) ---> Configs :-
   config(E, Heap) ---> Configs_,
   composeConfigs(Configs_, (strict(E_), E_), Configs).
config(strict(lambda(_,_)), Heap) ---> [config(constr(success, []), Heap)].
config(strict(v(P)), Heap) ---> [config(constr(success, []), Heap)] :-
   lookupBinding(Heap, v(P), nil).
config(strict(constr(_,[])), Heap) ---> [config(constr(success, []), Heap)] :- !.
config(strict(constr(_,Ps)), Heap) ---> [config(E, Heap)] :-
   propagateStrictness(Ps, E).





narrow(_, [], _, []).
narrow(P,[(constr(C,Xs),E)|Alts],Heap,[config(E_,Heap_)|Configs]) :-
   length(Xs,N),
   newVars(N,Qs),
   initializeVars(Heap, Qs, Heap__),
   addBinding(Heap__, [P, constr(C,Qs)], Heap_),
   zip(Xs,Qs,Subst),
   substitute(E,Subst,E_),
   narrow(P,Alts,Heap,Configs).

propagateEquality([P],[Q], P =:= Q).
propagateEquality([P|Ps],[Q|Qs], (P =:= Q) & E) :-
   propagateEquality(Ps,Qs,E).

propagateStrictness([P], strict(P)).
propagateStrictness([P|Ps], strict(P) & E) :-
   propagateStrictness(Ps, E).
