% -*- prolog -*-



substitute(v(X), Subst, v(X)) :-
   myMember([v(X),v(X)],Subst), !.
substitute(v(X), Subst, E) :-
   myMember([v(X),E],Subst).
substitute(v(X), Subst, v(X)) :-
   \+ myMember([v(X),_], Subst).

substitute(lambda(X,E), Subst, lambda(X,E_)) :-
   reduceDomain(Subst, [X], SubstRed),
   substitute(E,SubstRed,E_).

substitute(constr(C,Xs), Subst, constr(C,Xs_)) :-
   substituteList(Xs,Subst,Xs_).

substitute(let(Defs,E),Subst,let(Defs_,E_)) :-
   zip(Xs,Es,Defs),
   reduceDomain(Subst,Xs,SubstRed),
   substituteList(Es,SubstRed,Es_),
   zip(Xs,Es_,Defs_),
   substitute(E,SubstRed,E_).

substitute(letfree(X, E), Subst, letfree(X, E_)) :-
   reduceDomain(Subst,[X],SubstRed),
   substitute(E,SubstRed,E_).

substitute(app(E,X), Subst, app(E_,X_)) :-
   substitute(E,Subst,E_),
   substitute(X,Subst,X_).


substitute(case(A,E,Alts), Subst, case(A,E_,Alts_)) :-
   substitute(E, Subst, E_),
   substituteAlts(Alts, Subst, Alts_).

substitute(G=>E,Subst,G_=>E_) :-
   substitute(G, Subst, G_),
   substitute(E, Subst, E_).

substitute(E1 & E2, Subst, E1_ & E2_) :-
   substitute(E1, Subst, E1_),
   substitute(E2, Subst, E2_).

substitute(E1 =:= E2, Subst, E1_ =:= E2_) :-
   substitute(E1, Subst, E1_),
   substitute(E2, Subst, E2_).

substitute(E1 or E2, Subst, E1_ or E2_) :-
   substitute(E1, Subst, E1_),
   substitute(E2, Subst, E2_).

substitute(strict(E), Subst, strict(E_)) :-
   substitute(E, Subst, E_).

substituteAlts([],_,[]).
substituteAlts([(constr(C,Xs),E)|Alts], Subst, [(constr(C,Xs),E_)|Alts_]) :-
   reduceDomain(Subst, Xs, SubstRed),
   substitute(E, SubstRed, E_),
   substituteAlts(Alts, Subst, Alts_).


substituteList([],_,[]).
substituteList([E|Es], Subst, [E_|Es_]) :-
   substitute(E,Subst,E_),
   substituteList(Es,Subst,Es_).

reduceDomain(Subst, [], Subst).
reduceDomain(Subst, [X|Xs], Subst_) :-
   myMember([X,E],Subst), !,
   myDelete(Subst,[X,E],Subst__),
   reduceDomain(Subst__, Xs, Subst_).
reduceDomain(Subst, [_|Xs], Subst_) :-
   reduceDomain(Subst, Xs, Subst_).


   
reduceSubst(Subst, Vars, Subst_) :-
   reduceSubstSub(Subst, Vars, Vars, Subst_).

reduceSubstSub(_,_,[],[]).
reduceSubstSub(Subst, AllVars, [P|Ps], [[P,E]|Subst_]) :-
   myMember([P,E],Subst),
   freeVars(E, F),
   deleteList(F, AllVars, New),
   myAppend(AllVars, New, AllVars_),
   myAppend(Ps, New, Ps_),
   reduceSubstSub(Subst, AllVars_, Ps_, Subst_).

substIter(Subst, SubstIter) :-
   substIterSub(Subst, Subst, SubstIter).

substIterSub([], _, []).
substIterSub([[P,E]|SubstRest], Subst, [[P,E_]|SubstIter]) :-
   substituteIter(E,Subst,E_),
   substIterSub(SubstRest, Subst, SubstIter).

substituteIter(E, Subst, E) :-
   substitute(E, Subst, E), !.
substituteIter(E, Subst, E_) :-
   substitute(E, Subst, E__),
   substituteIter(E__, Subst, E_).


