% -*- prolog -*-


% curry syntax
:- op(500, xfx, =>).
:- op(500, xfy, &).
:- op(400, xfy, =:=).
:- op(300, xfy, or).



composeConfigs([],_,[]).
composeConfigs([config(E,Heap)|Cs],(E_,X),[config(E_,Heap)|Cs_]) :-
   copy_term((E_, X), Proceed),
   X = E,
   composeConfigs(Cs, Proceed, Cs_).


allVars([],_,[]).
allVars([config(E,Heap)|Cs],P,[config(P,Heap_)|Cs_]) :-
   addBinding(Heap, [P,E], Heap_),
   allVars(Cs,P,Cs_).




whnf(lambda(_,_),_).
whnf(constr(_,_),_).
whnf(v(P),Heap) :-
   lookupBinding(Heap,v(P),nil).


freeVars(E, F) :-
   freeVarsSub(E, F_),
   removeDups(F_, F).

freeVarsSub(v(X), [v(X)]).

freeVarsSub(lambda(X,E), F) :-
   freeVarsSub(E, F_),
   myDelete(F_, X, F).

freeVarsSub(constr(_,Xs), Xs).

freeVarsSub(let(Defs,E), F) :-
   zip(Xs,Es,Defs),
   freeVarsSubList(Es, F_Es),
   freeVarsSub(E, F_E),
   myAppend(F_E, F_Es, F_),
   deleteList(F_, Xs, F).

freeVarsSub(letfree(X, E), F) :-
   freeVarsSub(E, F_),
   myDelete(F_,X,F).

freeVarsSub(app(E,X), [X|F]) :-
   freeVarsSub(E,F).

freeVarsSub(case(_,E,Alts), F) :-
   freeVarsSub(E, F_E),
   freeVarsSubAlts(Alts, F_Alts),
   myAppend(F_E, F_Alts, F).

freeVarsSub(G=>E,F) :-
   freeVarsSub(G, F_G),
   freeVarsSub(E, F_E),
   myAppend(F_E, F_G, F).

freeVarsSub(E1 & E2, F) :-
   freeVarsSub(E1, F1),
   freeVarsSub(E2, F2),
   myAppend(F1,F2,F).

freeVarsSub(E1 =:= E2, F) :-
   freeVarsSub(E1, F1),
   freeVarsSub(E2, F2),
   myAppend(F1,F2,F).

freeVarsSub(E1 or E2, F) :-
   freeVarsSub(E1, F1),
   freeVarsSub(E2, F2),
   myAppend(F1,F2,F).

freeVarsSub(strict(E), F) :-
   freeVarsSub(E, F).

freeVarsSubList([],[]).
freeVarsSubList([E|Es],F) :-
   freeVarsSub(E, F_E),
   freeVarsSubList(Es, F_Es),
   myAppend(F_E, F_Es, F).

freeVarsSubAlts([],[]).
freeVarsSubAlts([(constr(_,Xs),E)|Alts], F) :-
   freeVarsSub(E, F_E_),
   deleteList(F_E_, Xs, F_E),
   freeVarsSubAlts(Alts, F_Alts),
   myAppend(F_E, F_Alts, F).


showCTerm(v(X)) :- write(v(X)).
showCTerm(constr(function, []))   :- !, write('<<function>>').
showCTerm(constr(C, []))   :- !, write(C).
showCTerm(constr(C, Args)) :- write('('), write(C), write(' '), showArgs(Args), write(')').
showArgs([]).
showArgs([T]) :- !, showCTerm(T).
showArgs([T|Args]) :- showCTerm(T), write(' '), showArgs(Args).